'''
Created on 05-Jun-2019

@author: abhinav
'''
from ..connection_pool import use_connection_pool
from ..aws.push_tasks import push_task
from ..base import LOG_WARN


@push_task
@use_connection_pool(ses_client="ses")
def send_email(sender, to_list, subject, body_text=None, body_html=None, ses_client=None):
    """
    Send email.
    Note: The emails of sender and receiver should be verified.
    PARAMS
    @sender: sender's email, string
    @to: list of receipient emails eg ['a@b.com', 'c@d.com']
    @subject: subject of the email
    @body: body of the email
    """
    try:
        body_data = {}
        if(body_text):
            body_data['Text'] = {
                'Data': body_text,
                'Charset': 'UTF-8'
            }
        if(body_html):
            body_data['Html'] = {
                'Data': body_html,
                'Charset': 'UTF-8'
            }
        
        response = ses_client.send_email(
            Source=sender,
            Destination={
                'ToAddresses': to_list
            },
            Message={
                'Subject': {
                    'Data': subject,
                    'Charset': 'UTF-8'
                },
                'Body': body_data
            }
        )
        if response['ResponseMetadata']['HTTPStatusCode'] == 200:
            return response
        else:
            return None
    except Exception as ex:
        LOG_WARN('ses_error', data=str(ex))
        return None



#! -*- coding: utf-8 -*-
'''
dynamodb model fields
'''



import pickle
import decimal
from datetime import datetime, date, timedelta

import ujson as json
from .errors import FieldValidationException
from .helpers import str_time, str_to_time,  timestamp2date, smart_str, date2timestamp
from .expression import Expression
from .helpers import validate_dict_to_dynamo_storage,\
    validate_list_to_dynamo_storage


__all__ = ['Attribute', 'CharField', 'IntegerField', 'FloatField',
           'DateTimeField', 'DateField', 'TimeDeltaField', 'TimeField',
           'BooleanField', 'DictField', 'SetField', 'ListField']


class Attribute(Expression):
    """Defines an attribute of the model.
    The attribute accepts strings and are stored in DynamoDB as
    they are - strings.
    Options
        name         -- alternate name of the attribute. This will be used
                        as the key to use when interacting with DynamoDB.
        hash_key     -- is hash_key. Table primary key.
        range_key    -- is range_key. Table primary key.
        indexed      -- Index this attribute. Unindexed attributes cannot
                        be used in queries. Default: False.
        unique       -- validates the uniqueness of the value of the
                        attribute.
        validator    -- a callable that can validate the value of the
                        attribute.
        default      -- Initial value of the attribute.
        
        projections -- ["field_name"...] if indexed
    """
    use_decimal_types = False
    field_type = 'char'
    
    _is_secondary_hash_key = None
    _is_secondary_range_key = None
    

    def __init__(self,
                 name=None,
                 hash_key=False,
                 range_key=False,
                 indexed=False,
                 required=False,
                 validator=None,
                 unique=False,
                 default=None,
                 projections=None,
                 default_using_instance = None,
                 **kwargs):
        super(Attribute, self).__init__(**kwargs)
        self.name = name
        self.unique = unique
        self.indexed = indexed
        self.default = default
        self.required = required
        self.validator = validator
        self.hash_key = hash_key
        self.range_key = range_key
        self.projections = projections
        self.default_using_instance = default_using_instance

    def __get__(self, instance, owner):
        try:
            return getattr(instance, '_' + self.name)
        except AttributeError:
            if callable(self.default):
                default = self.default()
            elif(instance and callable(self.default_using_instance)):
                default = self.default_using_instance(instance)                
            else:
                default = self.default
            if not instance:
                return self
            self.__set__(instance, default)
            return default

    def __set__(self, instance, value):
        setattr(instance, '_' + self.name, value)

    def typecast_for_read(self, value):
        """Typecasts the value for reading from DynamoDB."""
        return value

    def typecast_for_storage(self, value):
        """Typecasts the value for storing to DynamoDB."""
        # default store unicode
        return smart_str(value)

    def value_type(self):
        return str

    def acceptable_types(self):
        return (str, dict, list, float, int, set)

    def validate(self, instance):
        val = getattr(instance, self.name)
        errors = []
        # type_validation
        if val is not None and not isinstance(val, self.acceptable_types()):
            errors.append((self.name, 'Bad type',))
        # validate first standard stuff
        if self.required:
            if val is None or not str(val).strip():
                errors.append((self.name, 'required'))
        # validate using validator
        if self.validator:
            r = self.validator(self.name, val)
            if r:
                errors.extend(r)
        if errors:
            raise FieldValidationException(errors)


class CharField(Attribute):

    def __init__(self, min_length=0, max_length=None, **kwargs):
        super(CharField, self).__init__(**kwargs)
        self.max_length = max_length
        self.min_length = min_length

    def typecast_for_read(self, value):
        if not value:
            return ''
        return smart_str(value)

    def typecast_for_storage(self, value):
        """Typecasts the value for storing to DynamoDB."""
        if value is None:
            return None
        try:
            return str(value)
        except UnicodeError:
            return value.decode('utf-8')

    def validate(self, instance):
        errors = []
        try:
            super(CharField, self).validate(instance)
        except FieldValidationException as err:
            errors.extend(err.errors)

        val = getattr(instance, self.name)

        if val:
            val_len = len(val)
            if ((self.hash_key and val_len > 2048) or
                    (self.range_key and val_len > 1024) or
                    (self.max_length and val_len > self.max_length)):
                errors.append((self.name, 'exceeds max length'))
            elif val_len < self.min_length:
                errors.append((self.name, 'exceeds min length'))
            else:
                # size 400k
                pass

        if errors:
            raise FieldValidationException(errors)


class IntegerField(Attribute):

    field_type = 'integer'

    def __init__(self, minimum=None, maximum=None, **kwargs):
        super(IntegerField, self).__init__(**kwargs)
        self.maximum = maximum
        self.minimum = minimum

    def typecast_for_read(self, value):
        if value is None:
            return None
        return int(value)

    def typecast_for_storage(self, value):
        if value is None:
            return 0
        return int(value)

    def value_type(self):
        return int

    def acceptable_types(self):
        return (int,)

    def validate(self, instance):
        errors = []
        try:
            super(IntegerField, self).validate(instance)
        except FieldValidationException as err:
            errors.extend(err.errors)

        val = getattr(instance, self.name)

        if val and not isinstance(val, self.acceptable_types()):
            errors.append((self.name, 'type error, need integer'))

        if self.minimum and val < self.minimum:
            errors.append((self.name, 'exceeds minimum'))
        elif self.maximum and val > self.maximum:
            errors.append((self.name, 'exceeds maximum'))

        if errors:
            raise FieldValidationException(errors)


class FloatField(Attribute):

    use_decimal_types = True
    field_type = 'float'

    def typecast_for_read(self, value):
        return float(value)

    def typecast_for_storage(self, value):
        if value is None:
            value = 0
        return decimal.Decimal(str(value))

    def value_type(self):
        return float

    def acceptable_types(self):
        return (int, float)

    def validate(self, instance):
        errors = []
        try:
            super(FloatField, self).validate(instance)
        except FieldValidationException as err:
            errors.extend(err.errors)

        val = getattr(instance, self.name)

        if val and not isinstance(val, self.acceptable_types()):
            errors.append((self.name, 'type error, need float'))

        if errors:
            raise FieldValidationException(errors)


class BooleanField(Attribute):

    field_type = 'bool'

    def typecast_for_read(self, value):
        return value

    def typecast_for_storage(self, value):
        if value is None:
            return False
        return value

    def value_type(self):
        return bool

    def acceptable_types(self):
        return self.value_type()


class DateTimeField(Attribute):

    field_type = 'datetime'

    def __init__(self, auto_now=False, auto_now_add=False, **kwargs):
        super(DateTimeField, self).__init__(**kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add

    def typecast_for_read(self, value):
        try:
            # We load as if the timestampe was naive
            # And gently override (ie: not convert) to the TZ to UTC
            dt = str_to_time(value)
            return dt
        except TypeError:
            return None
        except ValueError:
            return None

    def typecast_for_storage(self, value):
        if isinstance(value, (str, bytes)):
            try:
                value = str_to_time(value)
            except TypeError:
                return None
        if not isinstance(value, date):
            raise TypeError("%s should be date object, and not a %s" %
                            (self.name, type(value)))
        if value is None:
            return None
        return str_time(value)

    def value_type(self):
        return (datetime, str, bytes)

    def acceptable_types(self):
        return self.value_type()


class DateField(Attribute):

    field_type = 'date'

    def __init__(self, auto_now=False, auto_now_add=False, **kwargs):
        super(DateField, self).__init__(**kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add

    def typecast_for_read(self, value):
        try:
            # We load as if it is UTC time
            # And assign (ie: not convert) the UTC TimeZone
            dt = str_to_time(value)
            return dt
        except TypeError:
            return None
        except ValueError:
            return None

    def typecast_for_storage(self, value):
        if not isinstance(value, date):
            raise TypeError("%s should be date object, and not a %s" %
                            (self.name, type(value)))
        if value is None:
            return None
        return str_time(value)

    def value_type(self):
        return date

    def acceptable_types(self):
        return self.value_type()


class TimeField(Attribute):

    field_type = 'time'

    def __init__(self, auto_now=False, auto_now_add=False, **kwargs):
        super(TimeField, self).__init__(**kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add

    def typecast_for_read(self, value):
        try:
            # We load as if the timestampe was naive
            # And gently override (ie: not convert) to the TZ to UTC
            # dt = str_to_time(value)
            # return dt
            time = timestamp2date(value)
            return time
        except TypeError:
            return None
        except ValueError:
            return None

    def typecast_for_storage(self, value):
        if isinstance(value, decimal.Decimal):
            return value
        if not isinstance(value, self.value_type()):
            raise TypeError("%s should be date object, and not a %s" %
                            (self.name, type(value)))
        if value is None:
            return None
        t = date2timestamp(value)
        try:
            t = float(t)
        except TypeError:
            raise TypeError("%s should be date object, and not a %s" %
                            (self.name, type(value)))
        time = decimal.Decimal('%.6f' % t)
        return time

    def value_type(self):
        return (datetime, str, bytes, date)

    def acceptable_types(self):
        return self.value_type()


class TimeDeltaField(Attribute):

    field_type = 'timedeltal'

    def __init__(self, **kwargs):
        super(TimeDeltaField, self).__init__(**kwargs)

    if hasattr(timedelta, "totals_seconds"):
        def _total_seconds(self, td):
            return td.total_seconds
    else:
        def _total_seconds(self, td):
            return (td.microseconds + 0.0 +
                    (td.seconds + td.days * 24 * 3600) * 10 ** 6) / 10 ** 6

    def typecast_for_read(self, value):
        try:
            # We load as if it is UTC time
            if value is None:
                value = 0.
            td = timedelta(seconds=float(value))
            return td
        except TypeError:
            return None
        except ValueError:
            return None

    def typecast_for_storage(self, value):
        if not isinstance(value, timedelta):
            raise TypeError("%s should be timedelta object, and not a %s" %
                            (self.name, type(value)))
        if value is None:
            return None

        return "%d" % self._total_seconds(value)

    def value_type(self):
        return timedelta

    def acceptable_types(self):
        return self.value_type()


class DictField(Attribute):

    field_type = 'dict'

    def __init__(self, **kwargs):
        super(DictField, self).__init__(**kwargs)

    def typecast_for_read(self, value):#dynamo object to local
        if value == '':
            return {}
        value = json.loads(json.dumps(value))#decimals becomes float
        return value

    def typecast_for_storage(self, value):#local object to dynamo
        """Typecasts the value for storing to DynamoDB."""
        value = validate_dict_to_dynamo_storage(value)
        return value

    def validate(self, instance):
        errors = []
        try:
            super(DictField, self).validate(instance)
        except FieldValidationException as err:
            errors.extend(err.errors)

        val = getattr(instance, self.name)

        if val and not isinstance(val, dict):
            errors.append((self.name, 'must a dict'))

        if errors:
            raise FieldValidationException(errors)


class ListField(Attribute):

    field_type = 'list'

    def __init__(self, **kwargs):
        super(ListField, self).__init__(**kwargs)

    def typecast_for_read(self, value):
        if not value:
            return []
        return value

    def typecast_for_storage(self, value):
        """Typecasts the value for storing to DynamoDB."""
        value = validate_list_to_dynamo_storage(value) 
        return value

    def validate(self, instance):
        errors = []
        try:
            super(ListField, self).validate(instance)
        except FieldValidationException as err:
            errors.extend(err.errors)

        val = getattr(instance, self.name)

        if val and not isinstance(val, list):
            errors.append((self.name, 'must a list'))

        if errors:
            raise FieldValidationException(errors)


class SetField(Attribute):

    field_type = 'set'

    def __init__(self, **kwargs):
        super(SetField, self).__init__(**kwargs)

    def typecast_for_read(self, value):
        if not value:
            return set()
        else:
            value = set(value)
        return value

    def typecast_for_storage(self, value):
        """Typecasts the value for storing to DynamoDB."""
        if value is None:
            value = set()
        return value

    def validate(self, instance):
        errors = []
        try:
            super(SetField, self).validate(instance)
        except FieldValidationException as err:
            errors.extend(err.errors)

        val = getattr(instance, self.name)

        if val and not isinstance(val, set):
            errors.append((self.name, 'must a set'))
        
        if errors:
            raise FieldValidationException(errors)


Fields = (Attribute, CharField, IntegerField, FloatField,
          DateTimeField, DateField, TimeDeltaField, TimeField,
          BooleanField, ListField, DictField, SetField)



'''
Created on Jul 7, 2016

@author: abhinav
'''

import sys
VER = sys.version_info[0]
import base64
import hashlib
import socket, errno

_VALID_STATUS_CODES = [1000, 1001, 1002, 1003, 1007, 1008,
												1009, 1010, 1011, 3000, 3999, 4000, 4999]

HANDSHAKE_STR = (
	 "HTTP/1.1 101 Switching Protocols\r\n"
	 "Upgrade: WebSocket\r\n"
	 "Connection: Upgrade\r\n"
	 "Sec-WebSocket-Accept: %(acceptstr)s\r\n\r\n"
)

GUID_STR = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

STREAM = 0x0
TEXT = 0x1
BINARY = 0x2
CLOSE = 0x8
PING = 0x9
PONG = 0xA

HEADERB1 = 1
HEADERB2 = 3
LENGTHSHORT = 4
LENGTHLONG = 5
MASK = 6
PAYLOAD = 7

MAXHEADER = 65536
MAXPAYLOAD = 33554432


from collections import deque
import codecs
import struct


def _check_unicode(val):
		if VER >= 3:
				return isinstance(val, str)
		else:
				return isinstance(val, unicode)



class WebSocketServerHandler(object):

	 def __init__(self, sock, headers):
			self.client = sock
			self.handshaked = False

			self.fin = 0
			self.data = bytearray()
			self.opcode = 0
			self.hasmask = 0
			self.maskarray = None
			self.length = 0
			self.lengtharray = None
			self.index = 0
			self.request = None
			self.usingssl = False

			self.frag_start = False
			self.frag_type = BINARY
			self.frag_buffer = None
			self.frag_decoder = codecs.getincrementaldecoder('utf-8')(errors='strict')
			self.closed = False
			self.state = HEADERB1

			# restrict the size of header and payload for security reasons
			self.maxheader = MAXHEADER
			self.maxpayload = MAXPAYLOAD


			 #additional fields convinience
			self.conn_obj = None
		
	 def start_handling(self):
			 while(True):
						try:
							 self._handleData()
						except (socket.error, Exception) as ex:
							 self.handleClose(ex)
							 self.client.close()
							 break
						except socket.timeout as ex:
								if(not self.handleTimeout()):
										self.handleClose(ex)
										self.client.close()
										break


	 def handleTimeout(self):
			 return False

	 def handleMessage(self):
			"""
					Called when websocket frame is received.
					To access the frame data call self.data.
					If the frame is Text then self.data is a unicode object.
					If the frame is Binary then self.data is a bytearray object.
			"""
			pass

	 def handleConnected(self):
			"""
					Called when a websocket client connects to the server.
			"""
			pass

	 def handleClose(self , ex):
			"""
					Called when a websocket server gets a Close frame from a client.
			"""
			pass

	 def _handlePacket(self):
			if self.opcode == CLOSE:
				 pass
			elif self.opcode == STREAM:
				 pass
			elif self.opcode == TEXT:
				 pass
			elif self.opcode == BINARY:
				 pass
			elif self.opcode == PONG or self.opcode == PING:
				 if len(self.data) > 125:
						raise Exception('control frame length can not be > 125')
			else:
					# unknown or reserved opcode so just close
				 raise Exception('unknown opcode')

			if self.opcode == CLOSE:
				 status = 1000
				 reason = u''
				 length = len(self.data)

				 if length == 0:
						pass
				 elif length >= 2:
						status = struct.unpack_from('!H', self.data[:2])[0]
						reason = self.data[2:]

						if status not in _VALID_STATUS_CODES:
								status = 1002

						if len(reason) > 0:
								try:
										reason = reason.decode('utf8', errors='strict')
								except:
										status = 1002
				 else:
						status = 1002

				 self.close(status, reason)
				 return

			elif self.fin == 0:
					if self.opcode != STREAM:
							if self.opcode == PING or self.opcode == PONG:
									raise Exception('control messages can not be fragmented')

							self.frag_type = self.opcode
							self.frag_start = True
							self.frag_decoder.reset()

							if self.frag_type == TEXT:
									self.frag_buffer = []
									utf_str = self.frag_decoder.decode(self.data, final=False)
									if utf_str:
											self.frag_buffer.append(utf_str)
							else:
									self.frag_buffer = bytearray()
									self.frag_buffer.extend(self.data)

					else:
							if self.frag_start is False:
									raise Exception('fragmentation protocol error')

							if self.frag_type == TEXT:
									utf_str = self.frag_decoder.decode(self.data, final=False)
									if utf_str:
											self.frag_buffer.append(utf_str)
							else:
									self.frag_buffer.extend(self.data)

			else:
					if self.opcode == STREAM:
							if self.frag_start is False:
									raise Exception('fragmentation protocol error')

							if self.frag_type == TEXT:
									utf_str = self.frag_decoder.decode(self.data, final=True)
									self.frag_buffer.append(utf_str)
									self.data = u''.join(self.frag_buffer)
							else:
									self.frag_buffer.extend(self.data)
									self.data = self.frag_buffer

							self.handleMessage()

							self.frag_decoder.reset()
							self.frag_type = BINARY
							self.frag_start = False
							self.frag_buffer = None

					elif self.opcode == PING:
							self._sendMessage(False, PONG, self.data)

					elif self.opcode == PONG:
							pass

					else:
							if self.frag_start is True:
									raise Exception('fragmentation protocol error')

							if self.opcode == TEXT:
									try:
											self.data = self.data.decode('utf8', errors='strict')
									except Exception as exp:
											raise Exception('invalid utf-8 payload')

							self.handleMessage()


	 def do_handshake(self, headers):
		
			try:
				 key = headers['Sec-WebSocket-Key'].strip()
				 k = key+GUID_STR
				 k_s = base64.b64encode(hashlib.sha1(k).digest()).decode('utf-8')
				 hStr = HANDSHAKE_STR % {'acceptstr': k_s}
				 self._sendBuffer(hStr.encode('utf-8'))
				 self.handshaked = True
				 self.handleConnected()
				 
#          base64.b64encode(
#                 hashlib.sha1(key + self.GUID).digest())
			except Exception as e:
				 raise Exception('handshake failed: %s', str(e))
		

	 def _handleData(self):
			# do the HTTP header and handshake
				 data = self.client.recv(8192)
				 if not data:
						raise Exception("remote socket closed")

				 if VER >= 3:
						 for d in data:
								 self._parseMessage(d)
				 else:
						 for d in data:
								 self._parseMessage(ord(d))

	 def close(self, status=1000, reason=u''):
			 """
					Send Close frame to the client. The underlying socket is only closed
					when the client acknowledges the Close frame.
					status is the closing identifier.
					reason is the reason for the close.
				"""
			 try:
					if self.closed is False:
						close_msg = bytearray()
						close_msg.extend(struct.pack("!H", status))
						if _check_unicode(reason):
								close_msg.extend(reason.encode('utf-8'))
						else:
								close_msg.extend(reason)

						self._sendMessage(False, CLOSE, close_msg)

			 finally:
						self.closed = True


	 def _sendBuffer(self, buff):
			size = len(buff)
			tosend = size
			already_sent = 0

			while tosend > 0:
				 try:
						# i should be able to send a bytearray
						sent = self.client.send(buff[already_sent:])
						if sent == 0:
							 raise RuntimeError('socket connection broken')

						already_sent += sent
						tosend -= sent

				 except socket.error as e:
						# if we have full buffers then wait for them to drain and try again
						if e.errno in [errno.EAGAIN, errno.EWOULDBLOCK]:
							 self._sendBuffer(buff[already_sent:])
						else:
							 raise e

			return None

	 def sendFragmentStart(self, data):
			"""
					Send the start of a data fragment stream to a websocket client.
					Subsequent data should be sent using sendFragment().
					A fragment stream is completed when sendFragmentEnd() is called.
					If data is a unicode object then the frame is sent as Text.
					If the data is a bytearray object then the frame is sent as Binary.
			"""
			opcode = BINARY
			if _check_unicode(data):
				 opcode = TEXT
			self._sendMessage(True, opcode, data)

	 def sendFragment(self, data):
			"""
					see sendFragmentStart()
					If data is a unicode object then the frame is sent as Text.
					If the data is a bytearray object then the frame is sent as Binary.
			"""
			self._sendMessage(True, STREAM, data)

	 def sendFragmentEnd(self, data):
			"""
					see sendFragmentEnd()
					If data is a unicode object then the frame is sent as Text.
					If the data is a bytearray object then the frame is sent as Binary.
			"""
			self._sendMessage(False, STREAM, data)

	 def send(self, data, is_text=True):
			"""
					Send websocket data frame to the client.
					If data is a unicode object then the frame is sent as Text.
					If the data is a bytearray object then the frame is sent as Binary.
			"""
			opcode = BINARY
			if is_text or _check_unicode(data):
				 opcode = TEXT
			self._sendMessage(False, opcode, data)


	 def _sendMessage(self, fin, opcode, data):

				payload = bytearray()

				b1 = 0
				b2 = 0
				if fin is False:
					 b1 |= 0x80
				b1 |= opcode

				if _check_unicode(data):
					 data = data.encode('utf-8')

				length = len(data)
				payload.append(b1)

				if length <= 125:
					 b2 |= length
					 payload.append(b2)

				elif length >= 126 and length <= 65535:
					 b2 |= 126
					 payload.append(b2)
					 payload.extend(struct.pack("!H", length))

				else:
					 b2 |= 127
					 payload.append(b2)
					 payload.extend(struct.pack("!Q", length))

				if length > 0:
					 payload.extend(data)

				self._sendBuffer(payload)


	 def _parseMessage(self, byte):
			# read in the header
			if self.state == HEADERB1:

				 self.fin = byte & 0x80
				 self.opcode = byte & 0x0F
				 self.state = HEADERB2

				 self.index = 0
				 self.length = 0
				 self.lengtharray = bytearray()
				 self.data = bytearray()

				 rsv = byte & 0x70
				 if rsv != 0:
						raise Exception('RSV bit must be 0')

			elif self.state == HEADERB2:
				 mask = byte & 0x80
				 length = byte & 0x7F

				 if self.opcode == PING and length > 125:
						 raise Exception('ping packet is too large')

				 if mask == 128:
						self.hasmask = True
				 else:
						self.hasmask = False

				 if length <= 125:
						self.length = length

						# if we have a mask we must read it
						if self.hasmask is True:
							 self.maskarray = bytearray()
							 self.state = MASK
						else:
							 # if there is no mask and no payload we are done
							 if self.length <= 0:
									try:
										 self._handlePacket()
									finally:
										 self.state = self.HEADERB1
										 self.data = bytearray()

							 # we have no mask and some payload
							 else:
									# self.index = 0
									self.data = bytearray()
									self.state = PAYLOAD

				 elif length == 126:
						self.lengtharray = bytearray()
						self.state = LENGTHSHORT

				 elif length == 127:
						self.lengtharray = bytearray()
						self.state = LENGTHLONG


			elif self.state == LENGTHSHORT:
				 self.lengtharray.append(byte)

				 if len(self.lengtharray) > 2:
						raise Exception('short length exceeded allowable size')

				 if len(self.lengtharray) == 2:
						self.length = struct.unpack_from('!H', self.lengtharray)[0]

						if self.hasmask is True:
							 self.maskarray = bytearray()
							 self.state = MASK
						else:
							 # if there is no mask and no payload we are done
							 if self.length <= 0:
									try:
										 self._handlePacket()
									finally:
										 self.state = HEADERB1
										 self.data = bytearray()

							 # we have no mask and some payload
							 else:
									# self.index = 0
									self.data = bytearray()
									self.state = PAYLOAD

			elif self.state == LENGTHLONG:

				 self.lengtharray.append(byte)

				 if len(self.lengtharray) > 8:
						raise Exception('long length exceeded allowable size')

				 if len(self.lengtharray) == 8:
						self.length = struct.unpack_from('!Q', self.lengtharray)[0]

						if self.hasmask is True:
							 self.maskarray = bytearray()
							 self.state = MASK
						else:
							 # if there is no mask and no payload we are done
							 if self.length <= 0:
									try:
										 self._handlePacket()
									finally:
										 self.state = HEADERB1
										 self.data = bytearray()

							 # we have no mask and some payload
							 else:
									# self.index = 0
									self.data = bytearray()
									self.state = PAYLOAD

			# MASK STATE
			elif self.state == MASK:
				 self.maskarray.append(byte)

				 if len(self.maskarray) > 4:
						raise Exception('mask exceeded allowable size')

				 if len(self.maskarray) == 4:
						# if there is no mask and no payload we are done
						if self.length <= 0:
							 try:
									self._handlePacket()
							 finally:
									self.state = HEADERB1
									self.data = bytearray()

						# we have no mask and some payload
						else:
							 # self.index = 0
							 self.data = bytearray()
							 self.state = PAYLOAD

			# PAYLOAD STATE
			elif self.state == PAYLOAD:
				 if self.hasmask is True:
						self.data.append(byte ^ self.maskarray[self.index % 4])
				 else:
						self.data.append(byte)

				 # if length exceeds allowable size then we except and remove the connection
				 if len(self.data) >= self.maxpayload:
						raise Exception('payload exceeded allowable size')

				 # check if we have processed length bytes; if so we are done
				 if (self.index + 1) == self.length:
						try:
							 self._handlePacket()
						finally:
							 # self.index = 0
							 self.state = HEADERB1
							 self.data = bytearray()
				 else:
						self.index += 1



'''
Created on 19-Jan-2018

@author: abhinav
'''


import urllib3
import json
from datetime import datetime
from elasticsearch import Elasticsearch, RequestsHttpConnection
from requests_aws4auth.aws4auth import AWS4Auth
from .config import aws_config, es_aws_host, es_http_host, IS_DEV
from .common_funcs_and_datastructures import cur_ms
urllib3.disable_warnings()

es_conn = None


if(es_http_host):
    es_conn = Elasticsearch([es_http_host])

        
elif(es_aws_host):
    es_awsauth = AWS4Auth(aws_config["aws_access_key_id"] , aws_config["aws_secret_access_key"] , aws_config["region_name"], 'es')
    es_conn = Elasticsearch(
                hosts=[{'host': es_aws_host, 'port': 443}],
                http_auth=es_awsauth,
                use_ssl=True,
                verify_certs=False,
                connection_class=RequestsHttpConnection
    )

#es_indexes_to_create = [{"index": name, "mapping": {}, "config": {}}...]
def create_indexes_and_mappings(es_indexes_to_create, recreate_indexes=False):
    #create items index
    #create communities index
    for index_to_create in es_indexes_to_create:
        index = index_to_create["index"]
        index_config = index_to_create.get("config")
        if(not index_config):
            index_config = {
                "settings": {
                    "number_of_shards": 1,
                    "number_of_replicas": 1,
                    "index.requests.cache.enable": True
                }
            }

        index_config["mappings"] = {
            "properties" : index_to_create["mappings"]
        }

        if(recreate_indexes):
            if(not isinstance(recreate_indexes, list) or (index in recreate_indexes)):
                es_conn.indices.delete(index, ignore=[400, 404])


        ret = es_conn.indices.create(index=index, body=index_config, ignore=400)
        IS_DEV and print("es_index_create", datetime.now(), {"index": index, "status": ret})



#! -*- coding: utf-8 -*-


import copy
from blaster.connection_pool import use_connection_pool

from .table import Table
from .fields import Fields
from .errors import FieldValidationException
import ujson as json


class Paginator(object):

    def __init__(self, model_object):
        self.model_object = model_object

    @use_connection_pool(db="dynamodb")
    def query(self, db=None, **kwargs):
        '''
        response_iterator = paginator.paginate(
            TableName='string',
            IndexName='string',
            Select='ALL_ATTRIBUTES'|'ALL_PROJECTED_ATTRIBUTES'|'SPECIFIC_ATTRIBUTES'|'COUNT',
            ConsistentRead=True|False,
            ScanIndexForward=True|False,
            ReturnConsumedCapacity='INDEXES'|'TOTAL'|'NONE',
            ProjectionExpression='string',
            FilterExpression='string',
            KeyConditionExpression='string',
            ExpressionAttributeNames={
                'string': 'string'
            },
            ExpressionAttributeValues={
                'string': {
                    'S': 'string',
                    'N': 'string',
                    'B': b'bytes',
                    'SS': [
                        'string',
                    ],
                    'NS': [
                        'string',
                    ],
                    'BS': [
                        b'bytes',
                    ],
                    'M': {
                        'string': {'... recursive ...'}
                    },
                    'L': [
                        {'... recursive ...'},
                    ],
                    'NULL': True|False,
                    'BOOL': True|False
                }
            },
            PaginationConfig={
                'MaxItems': 123,
                'PageSize': 123,
                'StartingToken': 'string'
            }
        )
        '''
        client = db.meta.client
        paginator = client.get_paginator('query')
        limit = kwargs.pop('Limit', None)
        kwargs['TableName'] = self.model_object.__table_name__
        if limit is not None:
            pagination_config = {
                'MaxItems': limit,
                'PageSize': limit,
                'StaringToken': '123232'
            }
            kwargs['PaginationConfig'] = pagination_config
        # response = paginator.paginate(**kwargs).build_full_result()
        response = paginator.paginate(**kwargs)
        for item in response:
            pass
        return item


class Query(object):

    def __init__(self, model_object, *args, **kwargs):
        self.Scan = False
        self.model_object = model_object
        self.model_class = self.model_object.__class__
        self.instance = self.model_class()
        self.ProjectionExpression = self._projection_expression(*args)
        self.ReturnConsumedCapacity = 'TOTAL'  # 'INDEXES'|'TOTAL'|'NONE'
        self.ConsistentRead = False
        self.FilterExpression = None
        self.ExclusiveStartKey = None  # 起始查询的key，就是上一页的最后一条数据
        self.KeyConditionExpression = None
        self.ExpressionAttributeNames = {}
        self.ExpressionAttributeValues = {}
        self.ScanIndexForward = False    # True|False
        self.ConditionalOperator = None  # 'AND'|'OR'
        self.IndexName = None
        self.Select = 'ALL_ATTRIBUTES'  # 'ALL_ATTRIBUTES'|'ALL_PROJECTED_ATTRIBUTES'|'SPECIFIC_ATTRIBUTES'|'COUNT'
        self.Offset = None
        self.Limit = None
        self.scaned_count = 0
        self.count = 0
        self.request_params = {}
        self.filter_args = []   # filter expression args
        self.filter_index_field = None  # index field name
        self.use_index_type = 0 #0=>primary, 1=>local secondary, 2=>global

    @property
    def consistent(self):
        self.ConsistentRead = True
        return self

    @property
    def scan(self):
        self.Scan = True
        return self

    def start_key_str(self, _str):
        if(_str):
            self.ExclusiveStartKey = json.loads(_str)
            self.request_params['ExclusiveStartKey'] = self.ExclusiveStartKey
        return self
    
    def start_key(self, cursor=None):
        if(cursor):
            self.ExclusiveStartKey = cursor
            self.request_params['ExclusiveStartKey'] = self.ExclusiveStartKey
        return self
    

    def _projection_expression(self, *args):
        instance = self.model_object
        projections = []
        for arg in args:
            if isinstance(arg, Fields):
                name = arg.name
                if arg not in instance.fields:
                    raise FieldValidationException('%s not found' % name)
                projections.append(name)
            else:
                raise FieldValidationException('Bad type must be Attribute type')
        ProjectionExpression = ",".join(projections)
        return ProjectionExpression

    def _get_primary_key(self):
        hash_key, range_key = self.instance._hash_key, self.instance._range_key
        key = {
            hash_key: getattr(self.instance, hash_key)
        }
        _range_key = getattr(self.instance, range_key, None)
        if range_key and not _range_key:
            raise FieldValidationException('Invalid range key value type')
        elif range_key:
            key[range_key] = _range_key
        return key

    def _filter_expression(self, *args):
        # get filter expression and key condition expression
        FilterExpression = None
        KeyConditionExpression = None
        params = {}
        for field_inst, exp, is_key in args:
            # field_inst = field_instance
            if field_inst.name == self.filter_index_field or \
                            (self.use_index_type==2  and  ( field_inst._is_secondary_hash_key or  field_inst._is_secondary_range_key )):
                _, exp, is_key = field_inst._expression_func(
                    field_inst.op, *field_inst.express_args, use_key=True)
            if is_key:
                if not KeyConditionExpression:
                    KeyConditionExpression = exp
                else:
                    KeyConditionExpression = KeyConditionExpression & exp
            else:
                if not FilterExpression:
                    FilterExpression = exp
                else:
                    FilterExpression = FilterExpression & exp
        if FilterExpression:
            params['FilterExpression'] = FilterExpression
        if KeyConditionExpression:
            params['KeyConditionExpression'] = KeyConditionExpression
        return params

    def _get_query_params(self):
        # get filter params
        params = {}
        # update filter expression
        filter_params = self._filter_expression(*self.filter_args)
        FilterExpression = filter_params.get('FilterExpression')
        if FilterExpression:
            params['FilterExpression'] = FilterExpression
        KeyConditionExpression = filter_params.get('KeyConditionExpression')
        if KeyConditionExpression:
            params['KeyConditionExpression'] = KeyConditionExpression
        if self.ProjectionExpression:
            params['ProjectionExpression'] = self.ProjectionExpression
        if self.ConsistentRead:
            params['ConsistentRead'] = self.ConsistentRead
        if self.ReturnConsumedCapacity:
            params['ReturnConsumedCapacity'] = self.ReturnConsumedCapacity
        self.request_params.update(params)
        return self.request_params

    def where(self, *args):
        # Find by any number of matching criteria... though presently only
        # "where" is supported.
        self.filter_args.extend(args)
        return self

    def limit(self, limit):
        self.Limit = limit
        self.request_params['Limit'] = limit
        return self

    def _get_item_params(self):
        params = {
            'Key': self._get_primary_key()
        }
        if self.ProjectionExpression:
            params['ProjectionExpression'] = self.ProjectionExpression
        if self.ConsistentRead:
            params['ConsistentRead'] = self.ConsistentRead
        return params

    @use_connection_pool(db="dynamodb")
    def get(self, db=None, **primary_key):
        # get directly by primary key
        self.instance = self.model_class(**primary_key)
        params = self._get_item_params()
        item = Table(self.instance, db).get_item(**params)
        if not item:
            return None
        value_for_read = self.instance._get_values_for_read(item)
        return value_for_read

    def first(self):
        response = self.limit(1).all()
        items = response['Items']
        return items[0] if items else None
    
    def use_index(self, index_field_or_name,asc=True):
        if isinstance(index_field_or_name, Fields):
            name = index_field_or_name.name
            index_name = self.instance._local_indexes.get(name)
            if not (index_name or index_field_or_name.range_key):
                raise FieldValidationException('index not found')
            self.filter_index_field = name
            if index_name:
                self.request_params['IndexName'] = index_name
            self.request_params['ScanIndexForward'] = asc
            self.use_index_type = 1
        elif isinstance(index_field_or_name, str):
            self.request_params['IndexName'] = index_field_or_name
            self.request_params['ScanIndexForward'] = asc
            index_name = self.model_class._global_indexes.get(index_field_or_name)
            if(index_name):
                self.use_index_type = 2
        elif(index_field_or_name==None):
            self.request_params['ScanIndexForward'] = asc
 
        return self

    def _yield_all(self, method, db):
        if method == 'scan':
            func = getattr(Table(self.instance, db), 'scan')
        elif method == 'query':
            func = getattr(Table(self.instance, db), 'query')
        else:
            return
        result_count = 0
        response = func(**self.request_params)
        while True:
            metadata = response.get('ResponseMetadata', {})
            for item in response['Items']:
                result_count += 1
                yield item
                if self.Limit > 0 and result_count >= self.Limit:
                    return
            LastEvaluatedKey = response.get('LastEvaluatedKey')
            if LastEvaluatedKey:
                self.request_params['ExclusiveStartKey'] = LastEvaluatedKey
                response = func(**self.request_params)
            else:
                break

    def _yield(self, db):
        if self.Scan:
            return self._yield_all('scan', db)
        else:
            return self._yield_all('query', db)

    @use_connection_pool(db="dynamodb")
    def all(self, db=None):
        self._get_query_params()
        if self.Scan:
            func = getattr(Table(self.instance, db), 'scan')
            #return self._yield_all('scan', db)
        else:
            func = getattr(Table(self.instance, db), 'query')
        response = func(**self.request_params)
        items = response['Items']
        results = []
        for item in items:
            _instance = self.model_class(**item)
            value_for_read = _instance._get_values_for_read(item)
            instance = self.model_class(**value_for_read)
            results.append(instance)
        response['Items'] = results
        return response
    
    def get_items(self, requery_for_all_projections=False):
        resp = self.all()
        items = resp["Items"]
        cursor = resp.get("LastEvaluatedKey", None)
        if(self.use_index_type != 0 and requery_for_all_projections and items):
            _items = self.model_class.batch_get(*map(lambda x : x.get_primary_key_dict() , items))
            #re-ordering back shit
            _items_map = {}
            for _item in _items:
                _items_map[_item.get_primary_key_tuple()] = _item
            
            items = [_items_map[item.get_primary_key_tuple()] for item in items]
        return items, cursor



_event_listeners = {} # {int : [listeners]}

PASS = NOT_HANDLING = object()

def register_as_listener(_id):
	if(callable(_id)):
		original_func = getattr(_id, "_original", _id)
		add_listener(original_func.__name__, _id)
		return _id
	else:
		def decorator(func):
			add_listener(_id, func)
			return func
		return decorator

# _n => max number of broadcasts, -1 => to all possible
def broadcast_event(_id, *args, _n=-1, **kwargs):
	listeners = _event_listeners.get(_id)
	if(not listeners):
		return
	listeners_ret_values = []
	set_arg_handled_count = "handled_count" in kwargs
	count = 0 # number of handlers handled this event, use case: not handling more than once
	for listener in listeners:
		if(set_arg_handled_count):
			kwargs["handled_count"] = count
		ret = listener(*args, **kwargs)
		if(ret != NOT_HANDLING):
			count += 1

		listeners_ret_values.append(ret)
		if(_n > 0 and _n <= count):
			break

	if(_n == 1): # special handling when you want just one response, don't return a tuple but just that value
		return listeners_ret_values and listeners_ret_values[0]

	#just fill upto _n values to return
	while(len(listeners_ret_values) < _n):
		listeners_ret_values.append(NOT_HANDLING)

	return tuple(listeners_ret_values)

def add_listener(_id, listener):
	print("registering_listener", "0", {"_id": _id, "listener": listener})
	
	if(not _id or not listener):
		return
	listeners = _event_listeners.get(_id)
	if(not listeners):
		_event_listeners[_id] = listeners = []

	listeners.insert(0, listener) # at the beginning

def remove_listener(_id, listener):
	listeners = _event_listeners.get(_id)
	if(not listeners):
		return
	if(listener in listeners):
		listeners.remove(listener)



import types
import pymysql
from itertools import chain
from databases import OrderedDict
import gevent
from gevent.queue import Queue

from .common_funcs_and_datastructures import jump_hash


class Attribute(object):
	_type = None

	def __init__(self, _type, **kwargs):
		self._type = _type
		self.__dict__.update(kwargs)

class Model(object):
	#class level
	_attrs = None
	_pk_attrs = None

	#instance level
	__is_new = True
	_pending_updates = None

	_inserted_id = None
	_init = False

	_json = None

	def get_custom_dict(self, path, dict_obj):

		_init = True

		class DictObj(dict):
			def __setitem__(this, k, v):
				if(not _init):
					new_path = path + "." + k
					self._pending_updates[new_path] = v
					if(isinstance(v, dict)):
						v = self.get_custom_dict(new_path, v)
				super(DictObj, this).__setitem__(k, v)

		ret = DictObj(**dict_obj)
		_init = False
		return ret

	def __setattr__(self, k, v):
		if(k in self.__class__._attrs):
			_attr_type_obj = self.__class__._attrs[k]
			if(_attr_type_obj._type == dict and isinstance(v, dict)):
				v = self.get_custom_dict(k, v)

			if(not self._init):
				self._pending_updates[k] = v
		self.__dict__[k] = v

	def to_json(self):
		self._json = _json = self._json or {}
		for k in self.__class__._attrs:
			if(k in self.__dict__):
				_json[k] = getattr(self, k , None)
		return _json

	def pk(self):
		_id = getattr(self, "_id", None)
		if(_id):
			return {"_id": _id}
		ret = {}
		for k in self.__class__._pk_attrs.keys():
			ret[k] = getattr(self, k)
		return ret

	#when retrieving objects from db
	@classmethod
	def get_instance_from_document(cls, doc):
		ret = cls()
		ret.__is_new = False
		ret._init = True
		ret._pending_updates = {}
		for k, v in cls._attrs.items():
			ret.__dict__[k] = None # set all initial attributes to None
		for k, v in doc.items():
			setattr(ret, k, v)
		ret._init = False
		return ret

	#used for creating new entities
	@classmethod
	def create_doc(cls, **values):
		ret = cls()
		ret._pending_updates = {}
		ret._pending_updates.update(cls.get_default_values())
		for k, v in cls._attrs.items():
			ret.__dict__[k] = None # set all initial attributes to None
		for k, v in values.items(): # set the given values
			setattr(ret, k, v)
		return ret

	@classmethod
	def get_default_values(cls):
		ret = {}
		for k, v in cls._attrs.items():
			default = getattr(v, "default", None)
			if(default):
				if(isinstance(default, types.FunctionType)):
					default = default()
				ret[k] = default
		return ret

	@classmethod
	def get_shard_key_from_query(cls, _query):
		_shard_keys = cls._shard_keys_
		if(not _shard_keys):
			return ""
		ret = []
		for k in cls._shard_keys_:
			val = _query.get(k)
			if(not val):
				return None
			ret.append(val) # must exist
		return "__".join(ret)

	def get_shard_key(self):
		_shard_keys = self.__class__._shard_keys_
		if(not _shard_keys):
			return ""
		ret = []
		for k in _shard_keys:
			ret.append(getattr(self, k))
		return "__".join(ret)

	@classmethod
	def get_databases_to_query(cls, _query, shard_keys_array):
		databases_to_update = None
		if(shard_keys_array):
			databases_set = {}
			for _shard_keys in shard_keys_array:
				_collection = None
				if(isinstance(_shard_keys, str)): # assuming single primary key
					_collection = get_collection(cls, _shard_keys)
				else:
					_collection = get_collection(cls, "__".join(_shard_keys))
				databases_set[id(_collection)] = _collection
			databases_to_update = databases_set.values()
		else:
			shard_key = cls.get_shard_key_from_query(_query)
			if(shard_key):
				databases_to_update = [get_collection(cls, shard_key)]
			else:
				#update in all collection
				databases_to_update = get_databases_all_nodes(cls)

		return databases_to_update

	@classmethod
	def update(cls, _query, update_values, shard_keys_array=None, **kwargs):
		databases_to_update = cls.get_databases_to_query(_query, shard_keys_array)

		multi_update_response = []

		def query_collection(_collection, _query, update_values):
			multi_update_response.append(_collection.update_many(_query, update_values, **kwargs))

		threads = [gevent.spawn(query_collection, _collection, _query, update_values)
													for _collection in databases_to_update
				]
		#waint on threads and return cursors
		gevent.joinall(threads)
		return multi_update_response

	@classmethod
	def query(cls, _query, sort=None, projection=None, shard_keys_array=None, **kwargs):
		databases_to_query = cls.get_databases_to_query(_query, shard_keys_array)

		multi_iterator = []

		def query_collection(_collection, _query, projection, sort):
			ret = _collection.find(_query, projection, **kwargs)
			if(sort):
				ret = ret.sort(sort)
			multi_iterator.append(ret)


		threads = [gevent.spawn(query_collection, _collection, _query, projection, sort)
													for _collection in databases_to_query
				]
		#waint on threads and return cursors
		gevent.joinall(threads)

		return map(cls.get_instance_from_document, chain(*multi_iterator))

	def commit(self, force=False):
		committed = False
		if(self.__is_new): # try inserting
			self.__is_new = False
			try:
				self._inserted_id = get_collection(self.__class__, self.get_shard_key()).insert_one(self._pending_updates)
				committed = True
			except DuplicateKeyError as ex:
				if(not force):
					raise(ex) # re reaise

				# try remove the primary key from pending updates
				if("_id" in self._pending_updates):
					del self._pending_updates["_id"]

				for k in list(self._pending_updates.keys()):
					if(k in self.__class__._pk_attrs):
						del self._pending_updates[k]

		if(not self.__is_new and not committed and self._pending_updates):  # try updated
			get_collection(self.__class__, self.get_shard_key()).update(self.pk(), {"$set": self._pending_updates}) # , hint=self.__class__._pk_attrs)

		self._pending_updates.clear()
		return self

	def delete(self):
		get_collection(self.__class__, self.get_shard_key()).delete_one(self.pk())



class GetCusror:

	def __init__(self, conn):
		self.conn = conn
		self.mysql_conn = None

	def __enter__(self):
		self.mysql_conn = self.conn.get_db()
		return self.mysql_conn

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.conn.release_db(self.mysql_conn)


'''
Single connection object abstracting pool of connections
'''
class Connection:

	db_args = None
	conn_pool = None

	def __init__(self, **db_args):
		self.db_args = db_args
		self.conn_pool = Queue()


	def get_db(self): # cache and get collection? needed?
		conn = self.conn_pool.get(block=False)
		if(not conn):
			conn = pymysql.connect(**self.db_args)
		return conn

	def release_db(self, conn): # cache and get collection? needed?
		self.conn_pool.put(conn)
	

	def inititalize(self, Model):
		if(isinstance(Model, list)):
			for _m in Model:
				self.inititalize(_m)
			return

		Model._attrs = _model_attrs = {}
		Model._pk_attrs = None

		attrs_to_name = {}
		for k, v in Model.__dict__.items():
			if(isinstance(v, Attribute)):
				_model_attrs[k] = v
				attrs_to_name[v] = k
				attrs_to_name[k] = k # dumb , but it's one time thing and also helps converting if any given attributes as strings

		# ensure indexes created and all databases loaded to memory
		_model_indexes = getattr(Model, "_index_", [])
		Model._shard_keys_ = getattr(Model, "_shard_keys_", ())
		Model._cluster_ = getattr(Model, "_cluster_", "default")

		_mysql_indexes_to_create = []
		for _index in _model_indexes:
			mysql_index = []
			if(not isinstance(_index, tuple)):
				_index = (_index,)
			for _a in _index:
				_attr_name = _a
				_ordering = 'ASC'
				if(isinstance(_a, tuple)):
					_a, _ordering = _a
				if(isinstance(_a, Attribute)):
					_attr_name = attrs_to_name[_a]

				mysql_index.append((_attr_name, _ordering))


			if(not Model._pk_attrs):
				Model._pk_attrs = _pk_attrs = OrderedDict()
				for i in mysql_index: # first unique index
					_pk_attrs[i[0]] = 1

			_mysql_indexes_to_create.append(mysql_index)

		# set collection name to include shard_keys
		if(Model._shard_keys_):
			Model._collection_name_ = Model._collection_name_ + "_shard_" + ("__".join(map(lambda x: attrs_to_name[x], Model._shard_keys_)))
		# create indices in mysql
		_index_sql = []
		_index_name = ""
		for mysql_index in _mysql_indexes_to_create:
			with GetCusror(self) as cursor:
				_sql = []
				for attr_name, order in mysql_index:
					_index_name += ("__" + attr_name)
					_sql.append("%s %s"%(attr_name, order))
			_index_sql.append("(%s)", ", ".join(_sql))

			cursor.execute("ALTER TABLE CREATE INDEX %s %s", (_index_name, _index_sql))


		if(not Model._pk_attrs): # create default _pk_attrs
			Model._pk_attrs = OrderedDict(_id=True)


''''
	cluster -> (node1 + node2 + node3) (each node could be replica set)
	data is client level sharded within node1, node2, node3
	queries are assembled as a chained iterator
'''
_mysql_clusters = {}
def init_mysql_cluster(nodes, cluster="default"):
	if(not isinstance(nodes, list)):
		nodes = [nodes]
	db_nodes = []
	for node in nodes:
		if(isinstance(node, dict)):
			node = Connection(**node)
		db_nodes.append(node)
	_mysql_clusters[cluster] = db_nodes


'''
	returns the collection from a perticular node depending on shard key
	shard_key=__1pk__2pk... underscore joined keys of a single primary key
'''
def get_collection(Model, shard_key):
	cluster_nodes = _mysql_clusters[Model._cluster_]
	_node_with_data = jump_hash(shard_key, len(cluster_nodes))
	return cluster_nodes[_node_with_data].get_collection(Model)

def get_databases_all_nodes(Model):
	cluster_nodes = _mysql_clusters[Model._cluster_]
	return map(lambda node: node.get_collection(Model), cluster_nodes)




# - Supports client-side sharding,
# - Doesn't autoshardi, you have to redistribute data if you add more db's later.

# - You can Configure mysql as a sharded cluster. Just pass the configuration server to 
# - init_mysql_cluster({"host": "", "port": 27017, "db_name": "yourdb", "user_name": "", "password": ""})



import re
from . import data_utils
PHONE_NUMBER_REGEX = re.compile("^\+?[0-9]+")


country_code_num_digits_map = {"93": [9], "358": [10], "355": [9], "213": [9], "1": [10], "374": [6], "297": [7], "61": [9], "672": [6], "43": [11], "994": [9], "973": [8], "880": [10], "375": [9], "32": [9], "501": [7], "229": [6, 7, 8, 9], "387": [8], "55": [11], "246": [7], "359": [9], "226": [8], "855": [9], "235": [8], "56": [9], "86": [11], "57": [10], "682": [5], "506": [8], "385": [9], "357": [8], "420": [9], "45": [8], "670": [8], "593": [9], "20": [10], "503": [7], "44": [10], "268": [8], "500": [5], "298": [5], "691": [7], "33": [9], "594": [9], "689": [6], "241": [7], "995": [9], "49": [10], "233": [9], "30": [10], "299": [6], "590": [12], "852": [8], "36": [9], "91": [10], "62": [10], "98": [10], "353": [9], "972": [9], "39": [10], "81": [11], "7": [10], "686": [5], "383": [8], "965": [8], "371": [8], "961": [8], "231": [8], "218": [10], "370": [8], "352": [9], "265": [7, 8, 9], "60": [7], "960": [7], "223": [8], "692": [7], "596": [12], "230": [8], "52": [10], "373": [8], "976": [8], "382": [8], "95": [10], "977": [10], "31": [9], "687": [6], "64": [8, 9, 10], "227": [8], "234": [8], "254": [9], "683": [4], "90": [11], "47": [8], "968": [8], "92": [10], "680": [7], "970": [9], "507": [8], "51": [9], "63": [10], "48": [9], "351": [9], "974": [8], "262": [12], "290": [4], "966": [9], "381": [8], "65": [8], "421": [9], "677": [7], "252": [7, 8], "27": [9], "34": [9], "94": [7], "46": [7], "41": [9], "963": [9], "886": [9], "66": [9], "228": [8], "216": [8], "380": [9], "971": [9], "58": [7], "84": [9], "967": [9]}

class SimpleTrie(dict):
	end_obj = None


country_code_num_digits_trie = SimpleTrie()


for i, j in country_code_num_digits_map.items():
	_root = country_code_num_digits_trie
	for k in i:
		if k not in _root:
			_root[k] = SimpleTrie()
		_root = _root[k]
	_root.end_obj = j



class PhoneNumberObj:
	phone_number = None
	country_phone_code = None
	national_number = None

	def __init__(self, country_phone_code, national_number, phone_number):
		self.phone_number = phone_number
		self.national_number = national_number
		if(country_phone_code):
			self.country_phone_code = country_phone_code.lstrip("+")


	#mobile numbers only
	@classmethod
	def parse_phone_number(cls, phone_number, country_phone_code=None, iso2_country_code=None):
		#assume phone number without country code
		#assume phone number with country code without +
		#assume + by mistake
		if(not phone_number or not PHONE_NUMBER_REGEX.match(phone_number)):
			return None
		
		if(country_phone_code):
			country_phone_code = country_phone_code.lstrip("+")

		phone_number = phone_number.lstrip("+0")

		_root = country_code_num_digits_trie
		for i in range(0, len(phone_number)):
			_d = phone_number[i]
			if(_root.end_obj):
				_remaining_len = len(phone_number) - i
				for _r in _root.end_obj:
					if(_r == _remaining_len):
						return PhoneNumberObj(phone_number[: i], phone_number[i:], phone_number)

			if(_d not in _root):
				break
			_root = _root[_d]


		if(not country_phone_code and iso2_country_code):
			country_phone_code = data_utils.iso2_to_phone_code.get(iso2_country_code, None) # netherlands by defailt

		if(not country_phone_code):
			country_phone_code = "31" # try netherlands by default


		if len(phone_number) in country_code_num_digits_map[country_phone_code]:
			return PhoneNumberObj(country_phone_code, phone_number, country_phone_code + phone_number)

		phone_number_strip = phone_number.lstrip(country_phone_code)
		if len(phone_number_strip) in country_code_num_digits_map[country_phone_code]:
			return PhoneNumberObj(country_phone_code, phone_number_strip, country_phone_code + phone_number_strip)

		if(phone_number.startswith(country_phone_code)):
			return PhoneNumberObj.parse_phone_number(phone_number.lstrip(country_phone_code), country_phone_code=country_phone_code, iso2_country_code=iso2_country_code)

		return None

############testing
def test():
	def assert_equals(a, b):
		if(a == b):
			print("yes")
		else:
			print("no")
	print(PhoneNumberObj.parse_phone_number('+917680971071').__dict__)
	print(PhoneNumberObj.parse_phone_number('+917680971071').__dict__)
	print(PhoneNumberObj.parse_phone_number("229270555", country_phone_code="31").__dict__)
	assert_equals(PhoneNumberObj.parse_phone_number('+917680971071', '+91').phone_number, "917680971071")
	assert_equals(PhoneNumberObj.parse_phone_number('917680971071', '+91').phone_number, "917680971071")
	assert_equals(PhoneNumberObj.parse_phone_number('07680971071', '+91').phone_number, "917680971071")
	assert_equals(PhoneNumberObj.parse_phone_number('7680971071', '+91').phone_number, "917680971071")
	assert_equals(PhoneNumberObj.parse_phone_number('7680971071', '+64').phone_number, "647680971071")
	assert_equals(PhoneNumberObj.parse_phone_number('76809710', '+64').phone_number, "6476809710")




def test_sanitized_dicts_list():
	from blaster.common_funcs_and_datastructures import SanitizedDict, SanitizedList
	sd = SanitizedDict(a="<a>", b="<b>")
	sd["c"] = "<c>"
	sd["d"] = {"e": "<e>", "f": "<f>"}

	for k, v in sd.items():
		print(k, v)
	for k, v in sd["d"].items():
		print(k, v)

	sl = SanitizedList(["<a>", "<b>"])
	sl.append({"c": "<c>", "d": "<d>"})
	sl.extend(["<e>", "<f>"])
	for i in sl:
		print(i)
	for k, v in sl[2].items():
		print(k, v)






def test():
	test_sanitized_dicts_list()


if __name__ == "__main__":
	test()



'''
Created on 27-Feb-2017

'''
import types
from datetime import datetime

import base64
import pickle
import traceback
import ujson as json
import gevent
from .. import config
from ..base import LOG_WARN, is_server_running
from ..common_funcs_and_datastructures import get_random_id,\
    cur_ms
from ..connection_pool import get_from_pool,\
    use_connection_pool
from ..base import base


push_tasks = {}
sqs_reader_greenlets = []

@use_connection_pool(queue="sqs")
def start_boto_sqs_readers(num_readers=5, msgs_per_batch=10, queue=None):
    if(not config.sqs_url):
        LOG_WARN("server_info", data="No sqs queue url provided, not starting readers")
        return
        
    queue_url = config.sqs_url
    
    def process_from_sqs():
        while(is_server_running()):
            try:
                response = queue.receive_message(
                        QueueUrl=queue_url,
                        MessageAttributeNames=['All'],
                        MaxNumberOfMessages=msgs_per_batch,
                        VisibilityTimeout=60,
                        WaitTimeSeconds=20 # long polling gevent safe
                        # ,ReceiveRequestAttemptId=''   , make it unique for each instance , probably when bootup with an ip ?
                )
                    
                for sqs_message in response.get("Messages", []):
                    message_payload = pickle.loads(
                        base64.a85decode(sqs_message.get("Body").encode())
                    )
                    kwargs = message_payload.get("kwargs", {})
                    args = message_payload.get("args", [])
                    func_name = message_payload.get("func_v2", "")
                    #task_id = message_payload.get("task_id", "")
                    #TODO use task_id for logging
                    func = push_tasks.get(func_name, None)
                    if func:
                        func(*args, **kwargs)
                    else:
                        LOG_WARN("server_exception", data="Not a push task", func=str(func))

                    _temp = queue.delete_message(
                        QueueUrl=queue_url,
                        ReceiptHandle=sqs_message.get("ReceiptHandle", None)
                    )
                    
                    LOG_WARN("sqs_processed", data=json.dumps(_temp))

            except Exception as ex:
                LOG_WARN("sqs_exception", stack_trace=traceback.format_exc())
        
    for i in range(num_readers):
        sqs_reader_greenlets.append(gevent.spawn(process_from_sqs))

    base.exit_listeners.append(wait_for_push_tasks_processing)


def push_task(func):
    if(isinstance(func, str)):
        #named push task
        def decorator(func2):
            push_tasks[func] = func2
            return func2

        return decorator
    else:
        #grab original function
        _original = getattr(func, "_original", func)
        push_tasks[_original.__name__] = func
        return func

@use_connection_pool(queue="sqs")
def post_a_task(func, *args, **kwargs):
    queue = kwargs.pop("queue")
    args = list(args)

    if isinstance(func, str) or isinstance(func, bytes):
        func_name = func
    elif isinstance(func, types.FunctionType):
        if hasattr(func, "_original"):
            func_name = func._original.__name__
        else:
            func_name = func.__name__
    else:
        return None

    now = datetime.utcnow().isoformat()
    task_id = get_random_id()
    message_body = {
        "args": args,
        "kwargs": kwargs,
        "func_v2": func_name,
        "task_id": task_id,
        "created_at": now
    }
    message_body = pickle.dumps(message_body)

    if(not config.sqs_url):
        message_body = pickle.loads(message_body)
        LOG_WARN("server_info", data="calling directly as not queue provided")
        func = push_tasks.get(
            message_body.get("func_v2", ""),
            None
        )
        func and func(*message_body.get("args", []), **message_body.get("kwargs", {}))
        return None

    response = queue.send_message(
        QueueUrl=config.sqs_url,
        MessageBody=base64.a85encode(message_body).decode(),# to utf-8 string
        DelaySeconds=1
    )
    return response
    


def wait_for_push_tasks_processing():
    LOG_WARN("server_info", data="wait called")
    gevent.joinall(sqs_reader_greenlets)
    del sqs_reader_greenlets[:]

def run_later(func):
    _original = getattr(func, "_original", func)
    task_name = _original.__name__
    push_tasks[task_name] = func

    def wrapper(*args, **kwargs):
        post_a_task(
            task_name,
            *args,
            **kwargs
        )

    wrapper._original = getattr(func, "_original", func)
    return wrapper



'''
Created on 22-Aug-2017

@author: abhinav
'''

# argparse==1.2.1
# gevent==1.2.1
# greenlet==0.4.12
# protobuf==3.2.0
# boto3==1.4.4

import ujson as json
import re
import urllib.parse
import time
import gevent
import signal
import functools
import traceback
import logging
from datetime import datetime
from gevent.server import StreamServer
from gevent.pywsgi import WSGIServer
from requests_toolbelt.multipart import decoder

from .. import config as blaster_config
from ..config import IS_DEV, DEBUG_LEVEL, LOG_LEVEL
from ..common_funcs_and_datastructures import cur_ms, SanitizedDict

_is_server_running = True

exit_listeners = []

default_wsgi_headers = [
	('Content-Type', 'text/html; charset=utf-8')
]
default_stream_headers = {
	'Content-Type': 'text/html; charset=utf-8',
	'Strict-Transport-Security': 'max-age=31536000; includeSubDomains',
	'X-Content-Type-Options': 'nosniff',
	'vary': 'Accept-Encoding',
	'X-Frame-Options': 'SAMEORIGIN',
	'X-UA-Compatible': 'IE=Edge',
	'X-XSS-Protection': '0'
}

I_AM_HANDLING_THE_SOCKET = object()
I_AM_HANDLING_THE_HEADERS = object()
I_AM_HANDLING_THE_STATUS = object()

REUSE_SOCKET_FOR_HTTP = object()

########some random constants
HTTP_MAX_REQUEST_BODY_SIZE = 1 * 1024 * 1024 # 1 mb
HTTP_MAX_HEADERS_DATA_SIZE = 16 * 1024 # 16kb

SLASH_N_ORDINAL = ord(b'\n')


def LOG(level, log_type, **kwargs):
	(level >= LOG_LEVEL) and print(log_type, datetime.now(), json.dumps(kwargs))

def LOG_PRINT(log_type, **kwargs):
	LOG(LOG_LEVEL, log_type, **kwargs)

def LOG_DEBUG(log_type, **kwargs):
	LOG(logging.DEBUG, log_type, **kwargs)

def LOG_INFO(log_type, **kwargs):
	LOG(logging.INFO, log_type, **kwargs)

def LOG_WARN(log_type, **kwargs):
	LOG(logging.WARN, log_type, **kwargs)

def LOG_ERROR(log_type, **kwargs):
	LOG(logging.ERROR, log_type, **kwargs)


def find_new_line(arr):
	n = len(arr)
	i = 0
	while(i < n):
		if(arr[i] == SLASH_N_ORDINAL):
			return i
		i += 1
	return -1


class BufferedSocket():
	
	is_eof = False
	sock = None
	store = None

	def __init__(self, sock):
		self.sock = sock
		self.store = bytearray()
		
	def close(self):
		self.sock.close()
		self.is_eof = True

	def send(self, *_data):
		for data in _data:
			if(isinstance(data, str)):
				data = data.encode()
			n = 0
			data_len = len(data)
			while(n < data_len):
				sent = self.sock.send(data[n:])
				n += sent
				if(sent < 0):
					return False
		return True
				
	def recv(self, n):
		if(self.is_eof):
			return None
		while(len(self.store) < n):
			data = self.sock.recv(4096)
			if(not data):
				self.is_eof = True
				break
			self.store.extend(data)
		#return n bytes for now
		ret = self.store[:n]
		#set remaining to new store
		self.store = self.store[n:]
		return ret
		
	def readline(self, max_size=HTTP_MAX_REQUEST_BODY_SIZE):
		if(self.is_eof):
			return None
		#indicates the index in store where we find the new line
		line_found = find_new_line(self.store)
		#check in store
		while(line_found == -1 and len(self.store) < max_size):
			data = self.sock.recv(4096)
			if(not data):
				self.is_eof = True
				break

			line_found = find_new_line(data)
			if(line_found != -1):
				line_found = len(self.store) + line_found

			self.store.extend(data) # fetch more data

		ret = None
		if(line_found != -1):
			ret = self.store[:line_found + 1]
			self.store = self.store[line_found + 1 :]
		else: # EOF, return everything in buffer
			ret = self.store
			self.store = bytearray()
			self.is_eof = True
		return ret

	def __getattr__(self, key):
		ret = getattr(self.sock, key, _OBJ_END_)
		if(ret == _OBJ_END_):
			raise AttributeError()
		return ret

####parse query string
def parse_qs_modified(qs, keep_blank_values=True, strict_parsing=False):
	_dict = {}
	for name, value in urllib.parse.parse_qsl(qs, keep_blank_values, strict_parsing):
		if name.endswith("[]"):
			_values = _dict.get(name)
			if(_values == None):
				_dict[name] = _values = []
			elif(not isinstance(_values , list)): # make it a list of values
				_dict[name] = _values = [_values]
			_values.append(value)
		else:
			_dict[name] = value
	return _dict



class HeadersDict(dict):
	def __getitem__(self, k):
		return super().__getitem__(k.lower())

	def get(self, k, default=None):
		header_value = super().get(k.lower(), _OBJ_END_)
		if(header_value == _OBJ_END_):
			return default
		return header_value


_OBJ_END_ = object()
class RequestParams:
	sock = None
	_post = None
	_get = None
	_cookies = None
	_headers = None
	#cookies to send to client
	_cookies_to_set = None
	#url data, doesn't contain query string
	path = None

	def __init__(self, buffered_socket):
		self._get = SanitizedDict() # empty params by default
		self._headers = HeadersDict()
		self.sock = buffered_socket

	#searches in post and get
	def get(self, key, default=None, **kwargs):
		val = _OBJ_END_
		#try post params first
		if(self._post):
			val = self._post.get(key, default=_OBJ_END_, **kwargs)
		#try get param next
		if(val == _OBJ_END_ and self._get):
			val = self._get.get(key, default=_OBJ_END_, **kwargs)

		if(val == _OBJ_END_):
			return default
		return val

	def __setitem__(self, key, val):
		self._get[key] = val

	def __getitem__(self, key):
		ret = self.get(key, default=_OBJ_END_)
		if(ret == _OBJ_END_):
			raise LookupError()
		return ret

	def __getattr__(self, key):
		ret = self.get(key, default=_OBJ_END_)
		if(ret == _OBJ_END_):
			raise AttributeError()
		return ret

	def GET(self, key=None, **kwargs):
		if(key == None):
			return self._get
		return self._get.get(key, **kwargs)

	def POST(self, key=None, **kwargs):
		if(key == None):
			return self._post
		if(self._post):
			return self._post.get(key, **kwargs)
		return None

	def HEADERS(self, key=None, default=None):
		if(key == None):
			return self._headers
		return self._headers.get(key, default=default)

	def COOKIES(self, key, cookie_value=_OBJ_END_):
		ret = self._cookies.get(key)
		if(cookie_value != _OBJ_END_):
			if(self._cookies_to_set == None):
				self._cookies_to_set = {}
			if(isinstance(cookie_value, dict)):
				_cookie_val = str(cookie_value.pop("value", "1"))
				if(cookie_value):
					_cookie_val += ";"
					#other key=val atrributes of cookie
					for k, v in cookie_value.items():
						_cookie_val += (" " + k)
						if(v is not True): # if mentioned as True just the key is suffice like httpsOnly, secureOnly
							_cookie_val += "=%s;"%(str(v))
				cookie_value = _cookie_val
			if(isinstance(cookie_value, str)):
				self._cookies_to_set[key] = cookie_value
		return ret

	def parse_request_body(self, post_data_bytes, headers):
		if(not post_data_bytes):
			return None
		self._post = _post_params = SanitizedDict()
		content_type_header = headers.get("Content-Type", "")
		if(headers and content_type_header.startswith("multipart/form-data")):
			for part in decoder.MultipartDecoder(post_data_bytes, content_type_header).parts:
				content_disposition = part.headers.pop(b'Content-Disposition', None)
				if(not content_disposition):
					continue
				attrs = {}
				for i in content_disposition.split(b';'):
					key_val = i.split(b'=')
					if(len(key_val) == 2):
						key = key_val[0].strip()
						val = key_val[1].strip(b'"\'').decode()
						attrs[key] = val
				name = attrs.pop(b'name', _OBJ_END_)
				if(name != _OBJ_END_):
					#if no additional headers, and no other attributes,
					#it's just a plain input field
					if(not part.headers and not attrs):
						_post_params[name] = part.text
					else: # could be a file or other type of data, preserve all data
						_post_params[name] = {
							"attrs": attrs,
							"data": part.content,
							"headers": part.headers
						}
		else:
			post_data_str = post_data_bytes.decode('utf-8')
			if(post_data_str[0] == '{'
				or content_type_header == "application/json"
			):
				_post_params.update(json.loads(post_data_str))
			else: # urlencoded form
				_post_params.update(parse_qs_modified(post_data_str))

		return self


	def to_dict(self):
		return {"get": self._get, "post": self._post}

class BlasterResponse:
	data = None
	templates = None
	frame_template = None
	meta_tags = None

	def __init__(self, data, template=None,
			frame_template=None, templates=None, meta_tags=None
		):
		self.data = data
		self.frame_template = frame_template
		self.meta_tags = meta_tags
		self.templates = {"content": template}
		if(templates):
			self.templates.update(templates)


#contains all running apps
_all_apps = set()
'''
	@app.route('/index.html')
'''
class App:
	route_handlers = None
	request_handlers = None
	name = None
	stream_server = None

	def __init__(self, name=None):
		self.name = name
		self.route_handlers = []
		self.request_handlers = []

	def route(self,
			regex,
			methods=None,
			name='',
			description='',
			query_class=None,
			request_class=None,
			response_class=None,
			max_body_size=None
	):
		methods = methods or ("GET", "POST")
		if(isinstance(methods, str)):
			methods = (methods,)

		def _decorator(func):
			self.route_handlers.append({
				"regex": regex,
				"func": func,
				"methods": methods,
				"name": name,
				"description": description,
				"query_class": query_class,
				"request_class": request_class,
				"response_class": response_class,
				"max_body_size": max_body_size
			})
			#return func as if nothing's decorated! (so you can call function as is)
			return func

		return _decorator


	def handle_wsgi(self, environ, start_response):
		ret = None
		for regex, method_handler in self.request_handlers:
			args = regex.match(environ['PATH_INFO'])
			func = method_handler.get(environ["REQUEST_METHOD"].upper())
			if(args != None and func):
				kwargs = {}
				kwargs["request_params"] = urllib.parse.parse_qs(environ['QUERY_STRING'])

				fargs = args.groups()
				if(fargs):
					ret = func(None, *fargs , **kwargs)
				else:
					ret = func(None, **kwargs)
				
				break
			
		status = response_headers = body = None
		if(ret != None):
			if(isinstance(ret, tuple)):
				l = len(ret)
				if(l > 0):
					body = ret[-1]
				if(l > 1):
					response_headers = ret[-2]
				if(l > 2):
					status = ret[-3]
			else:
				body = ret
			
			response_headers = response_headers or default_wsgi_headers
			status = status or "200 OK"
			start_response(status, response_headers)
			if(isinstance(body, list)):
				return body
			return [body]
		#close socket


	def start(self, port=80, handlers=[], use_wsgi=False, **ssl_args):
		self.route_handlers.sort(key=functools.cmp_to_key(lambda x, y: len(y["regex"]) - len(x["regex"]))) # reverse sort by length
		
		handlers_with_methods = []
		for handler in handlers:
			if(isinstance(handler, (list, tuple))):
				if(len(handler) < 2 or len(handler) > 3):
					continue
				handlers_with_methods.append({"regex": handler[0], "func": handler[1], "methods": ("GET",)})
			elif(isinstance(handler, dict)):
				_handler = {"methods": ("GET", "POST")} # defaults
				_handler.update(handler)
				handlers_with_methods.append(_handler)


		#add them to all route handlers
		self.route_handlers.extend(handlers_with_methods)

		regexes_map = {}
		#build {regex: {GET: handler,...}...}
		for handler in reversed(self.route_handlers):
			regex = handler["regex"]
			existing_regex_method_handlers = regexes_map.get(regex)
			if(not existing_regex_method_handlers):
				regexes_map[regex] = existing_regex_method_handlers = {}
				self.request_handlers.append((re.compile(regex), existing_regex_method_handlers))

			for method in handler["methods"]:
				existing_regex_method_handlers[method.upper()] = handler
		
		#sort ascending because we iterated in reverse earlier
		self.request_handlers.reverse()
		if(IS_DEV):
			for regex, method_handlers in self.request_handlers:
				print(regex.pattern, {method: handler["func"] for method, handler in method_handlers.items()})

		if(use_wsgi):
			self.stream_server = WSGIServer(('', port), self.handle_wsgi).serve_forever()
		else:
			class CustomStreamServer(StreamServer):
				def do_close(self, *args):
					return

				#lets handle closing ourself
				def wrap_socket_and_handle(self, client_socket, address):
					# used in case of ssl sockets
					return self.handle(self.wrap_socket(client_socket, **self.ssl_args), address)

			LOG_PRINT("server_start", port=port)
			self.stream_server = CustomStreamServer(('', port), handle=self.handle_connection, **ssl_args)
			#keep a track
		_all_apps.add(self)


	def serve(self):
		self.stream_server.serve_forever()

	def stop(self):
		self.stream_server.stop()
		_all_apps.remove(self)

	@classmethod
	def response_body_to_parts(cls, ret):
		status = response_headers = body = None
		if(isinstance(ret, tuple)):
			body_len = len(ret)
			if(body_len > 0):
				body = ret[-1]
			if(body_len > 1):
				response_headers = ret[-2]
			if(body_len > 2):
				status = str(ret[-3])
		else:
			body = ret

		return status, response_headers, body

	def process_http_request(self, buffered_socket):
		resuse_socket_for_next_http_request = True
		request_line = buffered_socket.readline(4096) # ignore request lines > 4096 bytes
		if(not request_line):
			return
		post_data = None
		request_params = RequestParams(buffered_socket)
		cur_millis = int(1000 * time.time())
		request_type = None
		request_path = None
		headers = None
		try:
			request_line = request_line.decode("utf-8")
			request_type, _request_line = request_line.split(" ", 1)
			_http_protocol_index = _request_line.rfind(" ")
			request_path \
				= request_params.path \
				= _request_line[: _http_protocol_index]
			#http_version = _request_line[_http_protocol_index + 1:]

			query_start_index = request_path.find("?")
			if(query_start_index != -1):
				query_string = request_path[query_start_index + 1:]
				request_params._get.update(parse_qs_modified(query_string))

				#strip query string from path
				request_path \
					= request_params.path \
					= request_path[:query_start_index]

			#find the handler
			handler = None
			for regex, method_handlers in self.request_handlers:
				args = regex.match(request_path)
				
				if(args != None):
					handler = method_handlers.get(request_type)
					if(not handler):
						#perform GET call by default
						LOG_PRINT(
							"handler_method_not_found",
							request_type=request_type,
							msg="performing GET by default"
						)
						if(request_type != "GET"):
							handler = method_handlers.get("GET")
					if(handler):
						#found a handler
						break

			if(not handler):
				raise Exception("Method not found")

			#parse the headers
			headers = request_params._headers
			max_headers_data_size = handler.get("max_headers_data_size") or HTTP_MAX_HEADERS_DATA_SIZE
			_headers_data_size = 0
			while(_headers_data_size < max_headers_data_size): # keep reading headers
				data = buffered_socket.readline(max_size=max_headers_data_size)
				if(data == b'\r\n'):
					break
				if(not data):
					return # won't resuse socket

				header_name_value_pair = data.split(b':', 1)
				if(len(header_name_value_pair) == 2):
					header_name , header_value = header_name_value_pair
					headers[header_name.lower().decode()] = header_value.lstrip(b' ').rstrip(b'\r\n').decode()
				
				_headers_data_size += len(data)


			#check if there is a content length or transfer encoding chunked
			content_length = int(headers.get("Content-Length", 0))
			_max_body_size = handler.get("max_body_size") or HTTP_MAX_REQUEST_BODY_SIZE
			if(content_length > 0):
				if(content_length < _max_body_size):
					post_data = buffered_socket.recv(content_length)

			else: # handle chuncked encoding
				transfer_encoding = headers.get("transfer-encoding")
				if(transfer_encoding == "chunked"):
					post_data = bytearray()
					chunk_size = int(buffered_socket.readline(), 16) # hex
					while(chunk_size > 0 and len(post_data) + chunk_size < _max_body_size):
						data = buffered_socket.recv(chunk_size + 2)
						post_data.extend(data)
						chunk_size = int(buffered_socket.readline(), 16) # hex
					#as bytes


			func = handler.get("func")
			#process cookies
			cookie_header = headers.get("Cookie", None)
			decoded_cookies = {}
			if(cookie_header):
				_temp = cookie_header.strip().split(";")
				for i in _temp:
					decoded_cookies.update(parse_qs_modified(i.strip()))
			request_params._cookies = decoded_cookies
			#process cookies end
			request_params.parse_request_body(post_data, headers)


			response_from_handler = func(*args.groups(), request_params=request_params)
			
			##app specific headers
			#handle various return values from handler functions
			(status, response_headers, body) = App.response_body_to_parts(response_from_handler)

			if(status != I_AM_HANDLING_THE_STATUS):
				status = status or '200 OK'
				buffered_socket.send(b'HTTP/1.1 ', status, b'\r\n')

			if(isinstance(response_headers, list)):
				#send only the specified headers
				for header in response_headers:
					buffered_socket.send(header, b'\r\n')

			else:
				if(isinstance(response_headers, dict)):
					#mix default stream headers
					#add default stream headers
					for key, val in default_stream_headers.items():
						if(key not in response_headers):
							buffered_socket.send(key, b': ', val, b'\r\n')

				elif(response_headers == None): # no headers => use default stream headers
					response_headers = default_stream_headers

				#send all basic headers
				for key, val in response_headers.items():
					buffered_socket.send(key, b': ', val, b'\r\n')

			#send any new cookies set
			if(request_params._cookies_to_set):
				for cookie_name, cookie_val in request_params._cookies_to_set.items():
					buffered_socket.send(b'Set-Cookie: ', cookie_name, b'=', cookie_val, b'\r\n')

			#check and respond to keep alive
			resuse_socket_for_next_http_request = headers.get('Connection', "").lower() == "keep-alive"
			#send back keep alive
			if(resuse_socket_for_next_http_request):
				buffered_socket.send(b'Connection: keep-alive\r\n')

			if(body == I_AM_HANDLING_THE_SOCKET):
				if(response_headers != I_AM_HANDLING_THE_HEADERS):
					buffered_socket.send(b'\r\n') # close the headers

				return I_AM_HANDLING_THE_SOCKET
			else:
				#just a variable to track api type responses
				response_data = None
				if(isinstance(body, BlasterResponse)):
					#just keep a reference as we overwrite body variable
					blaster_response_obj = body
					#set body as default to repose_obj.data
					body = response_data = blaster_response_obj.data
					return_type = request_params.get("return_type")

					#standard return_types -> json, content, None(=>content inside frame)
					if(return_type != "json"):
						#if template file is given and body should be a dict for the template
						templates = blaster_response_obj.templates
						#if no template given or returns just the blaster_response_obj.data
						template_exists \
							= (return_type and templates.get(return_type))\
							or blaster_response_obj.templates.get("content")

						if(template_exists and isinstance(response_data, dict)):
							tmpl_rendered = template_exists.render(
								request_params=request_params,
								**response_data
							)

							frame_template = blaster_response_obj.frame_template
							#if there is a return type, just retutn the template
							#render the full frame
							if(return_type or not frame_template):
								body = tmpl_rendered # partial

							else: # return type is nothing -> return full frame
								body = frame_template.render(
									body_content=tmpl_rendered,
									meta_tags=blaster_response_obj.meta_tags
								)

				if(isinstance(body, (dict, list))):
					response_data = body
					body = json.dumps(body)

				#encode body
				if(isinstance(body, str)):
					body = body.encode()

			#close headers and send the body data
			if(body):
				#finalize all the headers
				buffered_socket.send(b'Content-Length: ', str(len(body)), b'\r\n\r\n')
				buffered_socket.send(body)
			else:
				buffered_socket.send(b'Content-Length: 0', b'\r\n\r\n')

			#just some debug for apis
			if(IS_DEV and DEBUG_LEVEL > 1 and response_data):
				print(
					"##API DEBUG",
					json.dumps({
						"PATH": request_path,
						"REQUEST_METHOD": request_type,
						"GET": request_params._get,
						"REQUEST_BODY": request_params._post,
						"REQUEST_CONTENT_TYPE": headers.get("content-type"),
						"RESPONSE_STATUS": status,
						"RESPONSE_HEADERS": response_headers,
						"REPONSE_BODY": response_data
					}, indent=4)
				)

			LOG_PRINT("http", response_status=status, request_type=request_type , path=request_path, wallclockms=int(1000 * time.time()) - cur_millis)

		except Exception as ex:
			stacktrace_string = traceback.format_exc()
			body = None
			if(blaster_config.server_error_page):
				body = blaster_config.server_error_page(
					request_type,
					request_path,
					request_params,
					headers,
					stacktrace_string=stacktrace_string
				)
			if(not body):
				body = b'Internal server error'
			buffered_socket.send(
					b'HTTP/1.1 ', b'502 Server error', b'\r\n',
					b'Connection: close', b'\r\n',
					b'Content-Length: ', str(len(body)), b'\r\n\r\n',
					body
			)
			LOG_ERROR(
				"http_error",
				exception_str=str(ex),
				stacktrace_string=stacktrace_string,
				request_type=request_type,
				request_path=request_path,
				request_params=request_params.to_dict()
			)
			if(IS_DEV):
				traceback.print_exc()
			#BREAK THIS CONNECTION
		
		if(resuse_socket_for_next_http_request):
			return REUSE_SOCKET_FOR_HTTP

	'''
		all the connection handling magic happens here
	'''
	def handle_connection(self, socket, address):
		buffered_socket = BufferedSocket(socket)
		close_socket = True
		while(True):
			ret = self.process_http_request(buffered_socket)
			if(ret == REUSE_SOCKET_FOR_HTTP):
				continue # try again
			elif(ret == I_AM_HANDLING_THE_SOCKET):
				close_socket = False
			break

		if(close_socket):
			buffered_socket.close()


def stop_all_apps():
	LOG_PRINT("server_info", data="exiting all servers")
	global _is_server_running
	
	_is_server_running = False
	
	for i in exit_listeners:
		i()

	for app in list(_all_apps):
		app.stop() # should handle all connections gracefully

	#wait for queue to flush/process


gevent.signal_handler(signal.SIGINT, stop_all_apps)


# create a global app for generic single server use
_main_app = App()

#generic global route handler
route_handler = _main_app.route

#generic glonal route handler
def start_stream_server(*args, **kwargs):
	_main_app.start(*args, **kwargs)
	_main_app.serve()


def redirect_http_to_https():
	def redirect(path, request_params=None, user=None):
		host = request_params.HEADERS('host')
		if(not host):
			return "404", {}, "Not understood"
		return "302 Redirect", ["Location: https://%s%s?%s"%(host, request_params.path, request_params.query_string)], "Need Https"

	http_app = App("port=80")
	http_app.start(port=80, handlers=[("(.*)", redirect)])
	gevent.spawn(http_app.serve)
	return http_app


def is_server_running():
	return _is_server_running



import math
from math import cos, asin, sqrt


def distance(origin, destination):
    lat1, lon1 = origin
    lat2, lon2 = destination

    p = 0.017453292519943295     # Pi/180
    a = 0.5 - cos((lat2 - lat1) * p) / 2 + cos(lat1 * p) * cos(lat2 * p) * (1 - cos((lon2 - lon1) * p)) / 2
    return 12742 * asin(sqrt(a)) # 2*R*asin...



def distance_arc(origin, destination):
    lat1, lon1 = origin
    lat2, lon2 = destination

    # phi = 90 - latitude
    degrees_to_radians = 0.017453292519943295     # Pi/180
   
    phi1 = (90.0 - lat1) * degrees_to_radians
    phi2 = (90.0 - lat2) * degrees_to_radians
        
    # theta = longitude
    theta1 = lon1 * degrees_to_radians
    theta2 = lon2 * degrees_to_radians
        
    # Compute spherical distance from spherical coordinates.
        
    # For two locations in spherical coordinates
    # (1, theta, phi) and (1, theta, phi)
    # cosine( arc length ) =
    #    sin phi sin phi' cos(theta-theta') + cos phi cos phi'
    # distance = rho * arc length
    
    cos = (math.sin(phi1) * math.sin(phi2) * math.cos(theta1 - theta2) + math.cos(phi1) * math.cos(phi2))
    arc = math.acos(cos)

    return 6371 * arc



'''
Created on 29-Dec-2017

@author: abhinav
'''
from .model import Model
from .fields import CharField, IntegerField
from ..common_funcs_and_datastructures import cur_ms


class InboxMessage(Model):
    
    __table_name__ = 'inbox_messages'

    ReadCapacityUnits = 2
    WriteCapacityUnits = 2
    
    inbox_id = CharField(name="inbox_id", hash_key=True)
    created_at = IntegerField(name="created_at", range_key=True)
    
    #### Note dest_id => the owner who owns this inbox_id
    
    dest_session_id =  CharField(name="dest_session_id")
    src_id =  CharField(name="src_id", indexed=True)
    
    _type = IntegerField(name="_type")
    payload = CharField(name="payload")
    payload1 = CharField(name="payload1")
    payload2 = CharField(name="payload2")
    

    @classmethod
    def add_to_inbox(cls, inbox_id,  src_id, _type , payload, payload1=None, payload2 = None, created_at=None):
        if(created_at == None):
            created_at = cur_ms()
        payload = None if not payload else str(payload)
        payload1 = None if not payload1 else str(payload1)
        payload2 = None if not payload2 else str(payload2)

        return InboxMessage.create(inbox_id=inbox_id, src_id=src_id, _type=_type, payload=payload,  payload1=payload1, payload2=payload2, created_at=created_at  )
    
    @classmethod
    def add_to_inbox_multiple(cls, list_of_dicts ):
        for i in list_of_dicts:
            if(not i.get("created_at", None)):
                i["created_at"] = cur_ms()
        return InboxMessage.batch_write(list_of_dicts)
    
    
    
        

    @classmethod
    def get_inbox_messages(cls, inbox_id ,less_than_timestamp=None, greater_than_timestamp=None):
        if(greater_than_timestamp==None or greater_than_timestamp<0):
            greater_than_timestamp=0
        
        if(less_than_timestamp==None or less_than_timestamp<0):
            less_than_timestamp= cur_ms()
        items, cursor = InboxMessage.query().use_index(None, asc=False).where(InboxMessage.inbox_id.eq(inbox_id) ,  InboxMessage.created_at.between(int(greater_than_timestamp), int(less_than_timestamp))).get_items()
        if(items):
            less_than_timestamp = items[-1].created_at
            
        
        return items, less_than_timestamp, cursor!=None
    



# -*- coding: utf-8 -*-
"""
Python 富文本XSS过滤类
@package XssHtml
@version 0.1
@link http://phith0n.github.io/python-xss-filter
@since 20150407
@copyright (c) Phithon All Rights Reserved

Based on native Python module HTMLParser purifier of HTML, To Clear all javascript in html
You can use it in all python web framework
Written by Phithon <root@leavesongs.com> in 2015 and placed in the public domain.
phithon <root@leavesongs.com> 编写于20150407
From: XDSEC <www.xdsec.org> & 离别歌 <www.leavesongs.com>
GitHub Pages: https://github.com/phith0n/python-xss-filter
Usage:
    parser = XssHtml()
    parser.feed('<html code>')
    parser.close()
    html = parser.getHtml()
    print html

Requirements
Python 2.6+ or 3.2+
Cannot defense xss in browser which is belowed IE7
浏览器版本：IE7+ 或其他浏览器，无法防御IE6及以下版本浏览器中的XSS
"""
import re
try:
    from html.parser import HTMLParser
except:
    from HTMLParser import HTMLParser

class XssHtml(HTMLParser):
    allow_tags = ['a', 'img', 'br', 'strong', 'b', 'code', 'pre',
                  'p', 'div', 'em', 'span', 'h1', 'h2', 'h3', 'h4',
                  'h5', 'h6', 'blockquote', 'ul', 'ol', 'tr', 'th', 'td',
                  'hr', 'li', 'u', 'embed', 's', 'table', 'thead', 'tbody',
                  'caption', 'small', 'q', 'sup', 'sub']
    common_attrs = ["style", "class", "name"]
    nonend_tags = ["img", "hr", "br", "embed"]
    tags_own_attrs = {
        "img": ["src", "width", "height", "alt", "align"], 
        "a": ["href", "target", "rel", "title"],
        "embed": ["src", "width", "height", "type", "allowfullscreen", "loop", "play", "wmode", "menu"],
        "table": ["border", "cellpadding", "cellspacing"],
    }

    _regex_url = re.compile(r'^(http|https|ftp)://.*', re.I | re.S)
    _regex_style_1 = re.compile(r'(\\|&#|/\*|\*/)', re.I)
    _regex_style_2 = re.compile(r'e.*x.*p.*r.*e.*s.*s.*i.*o.*n', re.I | re.S)


    def __init__(self, allows=[]):
        HTMLParser.__init__(self)
        self.allow_tags = allows if allows else self.allow_tags
        self.result = []
        self.start = []
        self.data = []

    def getHtml(self):
        """
        Get the safe html code
        """
        for i in range(0, len(self.result)):
            self.data.append(self.result[i])
        return ''.join(self.data)

    def handle_startendtag(self, tag, attrs):
        self.handle_starttag(tag, attrs)

    def handle_starttag(self, tag, attrs):
        if tag not in self.allow_tags:
            return
        end_diagonal = ' /' if tag in self.nonend_tags else ''
        if not end_diagonal:
            self.start.append(tag)
        attdict = {}
        for attr in attrs:
            attdict[attr[0]] = attr[1]

        attdict = self._wash_attr(attdict, tag)
        if hasattr(self, "node_%s" % tag):
            attdict = getattr(self, "node_%s" % tag)(attdict)
        else:
            attdict = self.node_default(attdict)

        attrs = []
        for (key, value) in attdict.items():
            attrs.append('%s="%s"' % (key, self._htmlspecialchars(value)))
        attrs = (' ' + ' '.join(attrs)) if attrs else ''
        self.result.append('<' + tag + attrs + end_diagonal + '>')

    def handle_endtag(self, tag):
        if self.start and tag == self.start[len(self.start) - 1]:
            self.result.append('</' + tag + '>')
            self.start.pop()

    def handle_data(self, data):
        self.result.append(self._htmlspecialchars(data))

    def handle_entityref(self, name):
        if name.isalpha():
            self.result.append("&%s;" % name)

    def handle_charref(self, name):
        if name.isdigit():
            self.result.append("&#%s;" % name)

    def node_default(self, attrs):
        attrs = self._common_attr(attrs)
        return attrs

    def node_a(self, attrs):
        attrs = self._common_attr(attrs)
        attrs = self._get_link(attrs, "href")
        attrs = self._set_attr_default(attrs, "target", "_blank")
        attrs = self._limit_attr(attrs, {
            "target": ["_blank", "_self"]
        })
        return attrs

    def node_embed(self, attrs):
        attrs = self._common_attr(attrs)
        attrs = self._get_link(attrs, "src")
        attrs = self._limit_attr(attrs, {
            "type": ["application/x-shockwave-flash"],
            "wmode": ["transparent", "window", "opaque"],
            "play": ["true", "false"],
            "loop": ["true", "false"],
            "menu": ["true", "false"],
            "allowfullscreen": ["true", "false"]
        })
        attrs["allowscriptaccess"] = "never"
        attrs["allownetworking"] = "none"
        return attrs

    def _true_url(self, url):
        if self._regex_url.match(url):
            return url
        else:
            return "http://%s" % url

    def _true_style(self, style):
        if style:
            style = self._regex_style_1.sub('_', style)
            style = self._regex_style_2.sub('_', style)
        return style

    def _get_style(self, attrs):
        if "style" in attrs:
            attrs["style"] = self._true_style(attrs.get("style"))
        return attrs

    def _get_link(self, attrs, name):
        if name in attrs:
            attrs[name] = self._true_url(attrs[name])
        return attrs

    def _wash_attr(self, attrs, tag):
        if tag in self.tags_own_attrs:
            other = self.tags_own_attrs.get(tag)
        else:
            other = []

        _attrs = {}
        if attrs:
            for (key, value) in attrs.items():
                if key in self.common_attrs + other:
                    _attrs[key] = value
        return _attrs

    def _common_attr(self, attrs):
        attrs = self._get_style(attrs)
        return attrs

    def _set_attr_default(self, attrs, name, default=''):
        if name not in attrs:
            attrs[name] = default
        return attrs

    def _limit_attr(self, attrs, limit={}):
        for (key, value) in limit.items():
            if key in attrs and attrs[key] not in value:
                del attrs[key]
        return attrs

    def _htmlspecialchars(self, html):
        return html.replace("<", "&lt;")\
            .replace(">", "&gt;")\
            .replace('"', "&quot;")\
            .replace("'", "&#039;")


if "__main__" == __name__:
    parser = XssHtml()
    parser.feed("""<p><img src=1 onerror=alert(/xss/)></p><div class="left">
        <a href='javascript:prompt(1)'><br />hehe</a></div>
        <p id="test" onmouseover="alert(1)" style="expresSion(alert(1))">&gt;M<svg>
        <a href="https://www.baidu.com" target="self">MM</a></p>
        <embed src='javascript:alert(/hehe/)' allowscriptaccess=always />""")
    parser.close()
    print(parser.getHtml())


from setuptools import setup, find_packages

setup(
	name='blaster-server',
	packages=find_packages("."),
	version='0.0.355',
	license='MIT',
	description='Gevent based python server built from scratch',
	author='Abhinav Reddy',                   # Type in your name
	author_email='abhinavabcd@gmail.com',      # Type in your E-Mail
	url='https://github.com/abhinavabcd/blaster',
	download_url='https://github.com/abhinavabcd/blaster/archive/v0.0337b.tar.gz',
	keywords=['server', 'superfast', 'just like flask but faster and minimal'],
	install_requires=[            # I get to this in a second
		"wheel>=0.34.2",
		"boto3>=1.17.28",
		"botocore>=1.20.28",
		"pytz>=2020.1",
		"gevent>=20.9.0",
		"greenlet>=0.4.16",
		"pymongo>=3.10.1",
		"elasticsearch>=7.12.1",
		"ujson>=2.0.3",
		"python-dateutil>=2.8.1",
		"requests>=2.25.1",
		"requests-toolbelt>=0.9.1",
		"requests-aws4auth>=1.0.1",
		"PyMySQL>=0.9.3",
		"urllib3>=1.26.4"
	],
	classifiers=[
						'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
						'Intended Audience :: Developers',      # Define that your audience are developers
						'Topic :: Software Development :: Build Tools',
						'License :: OSI Approved :: MIT License',   # Again, pick a license
						'Programming Language :: Python :: 3',      # Specify which pyhton versions that you want to support
						'Programming Language :: Python :: 3.4',
						'Programming Language :: Python :: 3.5',
						'Programming Language :: Python :: 3.6',
	],
)



import tornado.httpserver
import tornado.ioloop
import tornado.options
import tornado.web

from tornado.options import define, options

define("port", default=8888, help="run on the given port", type=int)


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.write("Hello, world")


def main():
    tornado.options.parse_command_line()
    application = tornado.web.Application([(r"/test", MainHandler)])
    http_server = tornado.httpserver.HTTPServer(application)
    http_server.listen(options.port)
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    main()



'''
Created on 04-Nov-2017

@author: abhinav
'''
from os import environ
import logging

secrets = {
	"v0": "something_random_secret_here"
}

DEBUG_LEVEL = int(environ.get("DEBUG_LEVEL") or 1) # higher implies more debug information

IS_PROD = environ.get("IS_PROD") == "1"
#staging is for production test instances
IS_STAGING = IS_PROD and environ.get("IS_STAGING") == "1"
#
IS_DEV = 1 if not (IS_PROD or IS_STAGING) else 0

# CRITICAL-50 ERROR-40  WARNING-30  INFO-20  DEBUG-10  NOTSET-0
LOG_LEVEL = logging.DEBUG if IS_DEV else (logging.INFO if IS_STAGING else logging.WARN)

aws_config = {
	# 'aws_access_key_id': "A*****",
	# 'aws_secret_access_key': "G*****",
	# 'region_name': "ap-south-1"
}

es_http_host = None
es_aws_host = None

connection_generators = {}

sqs_url = None

server_error_page = None



import heapq
import gevent
import types
import pymongo
from bson.objectid import ObjectId
from itertools import chain
from collections import OrderedDict
from pymongo import MongoClient
from pymongo.errors import DuplicateKeyError
from pymongo import ReturnDocument, ReadPreference
from .common_funcs_and_datastructures import jump_hash, ExpiringCache, cur_ms, list_diff2, batched_iter
from .config import DEBUG_LEVEL as BLASTER_DEBUG_LEVEL, IS_DEV

MONGO_DEBUG_LEVEL = BLASTER_DEBUG_LEVEL
EVENT_BEFORE_DELETE = -2
EVENT_AFTER_DELETE = -1
EVENT_BEFORE_UPDATE = 1
EVENT_AFTER_UPDATE = 2
EVENT_BEFORE_CREATE = 3
EVENT_AFTER_CREATE = 4

#just a random constant object to check for non existent keys without catching KeyNotExist exception
_OBJ_END_ = object()
#given a list of [(filter, iter), ....] , flattens them

class Attribute(object):
	_type = None

	def __init__(self, _type, **kwargs):
		self._type = _type
		self.__dict__.update(kwargs)

class Model(object):
	#class level
	_attrs = None
	_pk_attrs = None

	__cache__ = ExpiringCache(10000)
	#instance level
	__is_new = True
	_set_query_updates = None
	_other_query_updates = None
	_original_doc = None
	#we execute all functions inside this list, and use the return value to update
	_insert_result = None
	#this means the object is being initialized by this orm, and not available to user yet
	_initializing = False
	_pk = None
	_json = None
	_is_secondary_shard = False # used to identify if it's a primary or secondary shard

	#initialize a new object
	def __init__(self, _is_new_=True, **values):

		#defaults
		self.__is_new = _is_new_
		self._initializing = not _is_new_
		self._set_query_updates = {}
		self._other_query_updates = {}
		cls = self.__class__
		for k, v in cls._attrs.items():
			self.__dict__[k] = None # set all initial attributes to None

		if(_is_new_):
			#setup default values
			self.__check_and_set_initial_defaults(values) # from empty doc

		#update given values
		for k, v in values.items():
			setattr(self, k, v)

	def get_id(self):
		return str(self._id)

	@classmethod
	def get(cls, _pk=None, use_the_cache=True, **kwargs):
		if(_pk != None):
			is_single_item = False
			_pks = None
			if(not isinstance(_pk, (list, set))):
				is_single_item = True
				_pks = [_pk]
			else:
				_pks = _pk

			if(not _pks):
				return []

			from_cache = []
			#handle _ids fetch separately so to speedup

			is_already_fetching = {}
			to_fetch_pks = []
			to_fetch_ids = []
			for _pk in _pks:
				_id = None # string based id
				_pk_tuple = None
				if(not _pk):
					continue
				elif(isinstance(_pk, str)):
					_pk_tuple = (_pk,)
					_id = _pk
					_pk = None # reset it to null as we are fetching it via ids
				else: # dict
					_pk_tuple = []
					for _k in cls._pk_attrs:
						_pk_tuple.append(_pk[_k])
					_pk_tuple = tuple(_pk_tuple)

				#check already in cache
				_item_in_cache = None
				if(use_the_cache and cls.__cache__):
					_item_in_cache = cls.__cache__.get(_pk_tuple)

				if(not _item_in_cache):
					if(is_already_fetching.get(_pk_tuple)):
						continue
					is_already_fetching[_pk_tuple] = True
					if(_pk):
						#remove duplicate pk query check if already fetching
						to_fetch_pks.append(_pk)
					if(_id):
						to_fetch_ids.append(_id)

				else:
					from_cache.append(_item_in_cache)

			_query = {"$or": to_fetch_pks}
			if(to_fetch_ids):
				to_fetch_pks.append({"_id": {"$in": to_fetch_ids}})

			ret = list(chain(from_cache, cls.query(_query)))
			if(is_single_item):
				if(ret):
					return ret[0]
				else:
					return None
			#multiple items, reorder
			pk_map = {}
			for item in ret:
				if(item):
					pk_map[item.pk_tuple()] = item
			id_map = {}
			for item in ret:
				if(item):
					id_map[item._id] = item
			ret = []
			for _pk in _pks:
				if(not _pk):
					continue
				elif(isinstance(_pk, str)): # possibly id
					item = id_map.get(_pk)
				else:
					_pk_tuple = []
					for _k in cls._pk_attrs:
						_pk_tuple.append(_pk.get(_k))
					_pk_tuple = tuple(_pk_tuple)

					item = pk_map.get(_pk_tuple)
				if(item):
					ret.append(item)

			return ret

		elif(kwargs):
			#assuming we want a single item
			ret = list(cls.query(kwargs))
			if(ret):
				return ret[0]

		return None

	'''Create a custom dict object, when you set an item on this dict, we basically mark pending_updates to mongo'''
	def get_custom_dict(self, path, _obj):

		_initializing = True

		class DictObj(dict):
			def __setitem__(this, k, v):
				if(not _initializing):
					new_path = path + "." + str(k)
					self._set_query_updates[new_path] = v
				if(_initializing):
					if(isinstance(v, dict)):
						v = self.get_custom_dict(new_path, v)
					elif(isinstance(v, list)):
						v = self.get_custom_list(new_path, v)

				super(DictObj, this).__setitem__(k, v)

			def pop(this, k, default=None):
				#not in initializing mode
				popped_val = super(DictObj, this).pop(k, _OBJ_END_)
				if(popped_val == _OBJ_END_):
					return default
				#if not initializing and has value set remove it form mongo
				elif(not _initializing):
					new_path = path + "." + str(k)
					_unset = self._other_query_updates.get("$unset")
					if(not _unset):
						_unset = self._other_query_updates["$unset"] = {}
					_unset[new_path] = ""
				#convert them to original object types
				if(isinstance(popped_val, list)):
					return list(popped_val)
				if(isinstance(popped_val, dict)):
					return dict(popped_val)
				return popped_val

			def update(self, another):
				for k, v in another.items():
					#calls __setitem__ again
					self[k] = v
				#allow chaining
				return self

		ret = DictObj(**_obj)
		_initializing = False
		return ret

	def get_custom_list(self, path, _list_obj):
		_initializing = True

		class ListObj(list):
			def __setitem__(this, k, v):
				if(not _initializing):
					new_path = path + "." + str(k)
					self._set_query_updates[new_path] = v
				if(_initializing):
					if(isinstance(v, dict)):
						v = self.get_custom_dict(new_path, v)
					elif(isinstance(v, list)):
						v = self.get_custom_list(new_path, v)
				super(ListObj, this).__setitem__(k, v)

			def remove(this, item): # can raise value exception
				super(ListObj, this).remove(item)
				# reset full array very inefficient : (
				self._set_query_updates[path] = this


			def pop(this, i=None):
				if(i == None):
					i = 1
					ret = super(ListObj, this).pop()
				else:
					ret = super(ListObj, this).pop(0)
					i = -1
				if(not _initializing):
					_pop = self._other_query_updates.get("$pop")
					if(_pop == None):
						_pop = self._other_query_updates["$pop"] = {}
					_pop[path] = i
				#convert them to original object types
				if(isinstance(ret, list)):
					return list(ret)
				if(isinstance(ret, dict)):
					return dict(ret)
				return ret

			def append(this, item):
				if(_initializing):
					if(isinstance(item, dict)):
						new_path = path + "." + str(len(this))
						item = self.get_custom_dict(new_path, item)
					elif(isinstance(item, list)):
						new_path = path + "." + str(len(this))
						item = self.get_custom_list(new_path, item)

				super(ListObj, this).append(item)

				if(not _initializing):
					_push = self._other_query_updates.get("$push")
					if(_push == None):
						_push = self._other_query_updates["$push"] = {}
					_push[path] = item

		ret = ListObj()
		for item in _list_obj:
			ret.append(item)
		_initializing = False
		return ret

	def __setattr__(self, k, v):
		_attr_type_obj = self.__class__._attrs.get(k)
		if(_attr_type_obj):
			#change type of objects when setting them
			if(v and not isinstance(v, _attr_type_obj._type)):
				if(_attr_type_obj._type in (int, float)): # force cast between int/float
					v = _attr_type_obj._type(v)
				elif(_attr_type_obj._type == str): # force cast to string
					v = str(v)
				elif(not self._initializing):
					raise Exception("Type mismatch in %s, %s: should be %s but got %s"%(type(self), k, _attr_type_obj._type, str(type(v))))

			if(self._initializing):
				if(not self.__is_new):
					if(_attr_type_obj._type == dict):
						if(isinstance(v, dict)):
							v = self.get_custom_dict(k, v)

					elif(_attr_type_obj._type == list):
						if(isinstance(v, list)):
							v = self.get_custom_list(k, v)
			else:
				cur_value = getattr(self, k, _OBJ_END_)
				if(cur_value != v):
					self._set_query_updates[k] = v
		self.__dict__[k] = v

	def to_dict(self):
		_json = {}
		for k in self.__class__._attrs:
			v = getattr(self, k, _OBJ_END_)
			if(v != _OBJ_END_):
				_json[k] = v

		_id = getattr(self, '_id', _OBJ_END_)
		if(_id != _OBJ_END_):
			_json['_id'] = str(_id)
		return _json


	@classmethod
	def pk_from_doc(cls, doc):
		ret = OrderedDict()
		for attr in cls._pk_attrs:
			ret[attr] = doc.get(attr)
		return ret

	def pk_tuple(self):
		return tuple(self.pk().values())

	def pk(self, renew=False):
		if(not self._pk or renew or self.__is_new):
			ret = OrderedDict()
			for k in self.__class__._pk_attrs.keys():
				ret[k] = getattr(self, k)
			self._pk = ret
		return self._pk

	#check all defaults and see if not present in doc, set them on the object
	def __check_and_set_initial_defaults(self, doc):
		for attr_name, attr_obj in self.__class__._attrs.items():
			_default = getattr(attr_obj, "default", _OBJ_END_)
			#if the value is not the document and have a default do the $set
			if(_default != _OBJ_END_ and attr_name not in doc):
				if(isinstance(_default, types.FunctionType)):
					_default = _default()
				#if there is a default value
				if(attr_obj._type == dict):
					self._set_query_updates[attr_name] = _default = dict(_default) # copy
				elif(attr_obj._type == list):
					self._set_query_updates[attr_name] = _default = list(_default) # copy
				else:
					self._set_query_updates[attr_name] = _default # copy

				setattr(self, attr_name, _default)


	#when retrieving objects from db
	@classmethod
	def get_instance_from_document(cls, doc):
		ret = cls(False) # not a new item

		#set defaults
		ret.__check_and_set_initial_defaults(doc)

		for k, v in doc.items():
			setattr(ret, k, v)

		ret._original_doc = doc
		#get and cache pk
		ret.pk(renew=True)
		#important flag to indicat initiazation finished
		ret._initializing = False
		if(cls.__cache__):
			cls.__cache__.set(ret.pk_tuple(), ret)
		return ret

	def reinitialize_from_doc(self, doc):
		cls = self.__class__
		#remove existing pk_tuple in cache
		cls.remove_from_cache(self)
		self._initializing = True

		self.__check_and_set_initial_defaults(doc)
		for k, v in doc.items():
			setattr(self, k, v)

		self._initializing = False
		self._original_doc = doc
		#renew the pk!
		self.pk(True)
		if(cls.__cache__):
			cls.__cache__.set(self.pk_tuple(), self)


	# basically finds the node where the shard_key resides
	# and returns a actual pymongo collection, shard_key must be a string
	@classmethod
	def get_collection(Model, shard_key):
		shard_key = str(shard_key) if shard_key != None else ""
		node_with_data_index = jump_hash(shard_key.encode(), len(Model._db_nodes_))
		db_node = Model._db_nodes_[node_with_data_index]
		return db_node.get_collection(Model)

	# returns all nodes and connections to Model inside them
	@classmethod
	def get_collection_on_all_db_nodes(Model):
		return map(lambda db_node: db_node.get_collection(Model), Model._db_nodes_)

	#give a shard key from query we return multiple connect
	@classmethod
	def get_collections_from_shard_key(cls, shard_key):
		if(isinstance(shard_key, (str, int))):
			return [cls.get_collection(shard_key)]
		elif(isinstance(shard_key, dict)):
			_in_values = shard_key.get("$in", _OBJ_END_)
			if(_in_values != _OBJ_END_):
				collections_to_update = {}
				#group all shards into one set
				for shard_key_value in _in_values:
					_collection = cls.get_collection(shard_key_value)
					collections_to_update[id(_collection)] = _collection
				return collections_to_update.values()

			_eq_value = shard_key.get("$eq", _OBJ_END_)
			if(_eq_value != _OBJ_END_):
				return [cls.get_collection(_eq_value)]

		elif(shard_key == None):
			IS_DEV and MONGO_DEBUG_LEVEL > 1 and print(
				"None type shard keys will be ignored and wont be available for query! Let me know feedback!"
			)
			return []

		raise Exception("Shard keys must be integers or strings or {'$in': [VALUES]}: got %s"%(str(shard_key),))

	@classmethod
	def get_collections_from_query(cls, _query):
		shard_key = _query.get(cls._shard_key_)

		if(shard_key):
			return cls.get_collections_from_shard_key(shard_key)

		#could'nt finds shards, return all shards
		return cls.get_collection_on_all_db_nodes()

	#_add_query is more query other than pk
	# for example, you can say update only when someother field > 0

	#force_update_secondary = [shard1, shard2]
	def update(self, _update_query, more_conditions=None, force_update_secondary=None, **kwargs):
		cls = self.__class__
		
		cls._trigger_event(EVENT_BEFORE_UPDATE, self)

		updated_doc = None
		if(_update_query):
			_query = dict(self.pk())
			if(more_conditions == None):
				more_conditions = {}

			#TODO: add existing values to more_conditions to be
			# super consistent on updates and probably
			# raise concurrent update exception if not modified
			_query.update(more_conditions)

			#get the shard where current object is
			primary_collection_shard = cls.get_collection(
				getattr(self, cls._shard_key_)
			)
			#query and update the document
			IS_DEV and MONGO_DEBUG_LEVEL > 1 and print(
				"#MONGO: update before and query",
				cls, _query, self._original_doc, _update_query
			)
			updated_doc = primary_collection_shard.find_one_and_update(
				_query,
				_update_query,
				return_document=ReturnDocument.AFTER,
				**kwargs
			)
			IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: after update::", cls, updated_doc)
			
		elif(force_update_secondary):
			IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: force updating secondary shards", cls, self.pk(), force_update_secondary)
			#we are foce updating secondary shards
			updated_doc = self._original_doc

		if(not updated_doc):
			#update was unsuccessful
			return None

		#CONS of this multi table sharding:
		#- Primary index updated by secondary constraint can fail
		#- in that case, the data exists in primary shard, but it's not valid data
		#TODO:
		# idea1: move this to transactions if we have secondary shards ?
		# idea2: revert the original update
		#update all secondary shards
		for shard_key, shard in cls._secondary_shards_.items():
			secondary_Model = shard._Model_
			#all secondary index is retrieved by _id, but sharded by their shard_key
			_secondary_pk = {"_id": self._original_doc["_id"]}
			_secondary_values_to_set = {}
			_secondary_values_to_unset = {}
			_secondary_collection = secondary_Model.get_collection(
				self._original_doc.get(shard_key)
			)

			_shard_key_changed = self._original_doc.get(shard_key) != updated_doc.get(shard_key)

			if(_shard_key_changed or (force_update_secondary and (shard_key in force_update_secondary))):
				#delete from old shard
				_secondary_collection.find_one_and_delete(_secondary_pk)
				for attr, _attr_obj in shard.attributes.items():
					new_value = updated_doc.get(attr)
					_secondary_values_to_set[attr] = new_value
				_secondary_values_to_set["_id"] = updated_doc["_id"]
				#find new shard
				_secondary_collection = secondary_Model.get_collection(
					_secondary_values_to_set[shard_key]
				)
				#insert new one
				IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: deleting and inserting new secondary",
					_secondary_pk, _secondary_values_to_set
				)
				_secondary_collection.insert_one(_secondary_values_to_set)
				
			else:
				#if someone else has modified
				#old values check in pk will not overwrite it
				for attr in shard.attributes:
					old_value = self._original_doc.get(attr, _OBJ_END_)
					new_value = updated_doc.get(attr, _OBJ_END_)
					if(old_value != new_value):
						if(new_value == _OBJ_END_): # removed after update
							_secondary_values_to_unset[attr] = new_value # unset this value as it doesn't exist anymore
						else:
							_secondary_values_to_set[attr] = new_value
						#also add this for query for a little bit of more consistency check in query
						if(old_value != _OBJ_END_): # means it exists
							_secondary_pk[attr] = old_value

				if(_secondary_values_to_set or _secondary_values_to_unset):
					_secondary_updates = {}
					if(_secondary_values_to_set):
						_secondary_updates["$set"] = _secondary_values_to_set
					if(_secondary_values_to_unset):
						_secondary_updates["$unset"] = _secondary_values_to_unset
					IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: updating secondary", _secondary_pk, _secondary_updates)
					#find that doc and update
					_secondary_collection.find_one_and_update(
						_secondary_pk,
						_secondary_updates
					)

		#reset all values
		self.reinitialize_from_doc(updated_doc)
		#waint on threads
		
		#triggered after update event
		cls._trigger_event(EVENT_AFTER_UPDATE, self)
		return True

	@classmethod
	def query(cls,
		_query,
		sort=None,
		projection=None,
		offset=None,
		limit=None,
		_no_requery=False,
		read_preference=ReadPreference.PRIMARY,
		**kwargs
	):
		queries = None
		# split or queries to separate it to shards
		if(not isinstance(_query, list)):
			if("$or" in _query):
				queries = _query["$or"]
			else:
				queries = [_query]
		else:
			queries = _query

		# if there is nothing to query return empty
		if(not queries):
			return []

		#normalize sort param to [('attr_name', ASCENDING/DESCENDING)...]
		if(sort):
			if(not isinstance(sort, list)):
				sort = [sort]
			_sort_keys = []
			for _sort_key_tuple in sort:
				if(not isinstance(_sort_key_tuple, tuple)):
					_sort_keys.append((cls._attrs_to_name[_sort_key_tuple], 1))
				else:
					_sort_keys.append((cls._attrs_to_name[_sort_key_tuple[0]], _sort_key_tuple[1]))
			sort = _sort_keys

		#map of collection: [_query]
		collections_to_query = {}
		for _query in queries:
			not_possible_in_the_secondary_shard = True
			shard_key = _query.get(cls._shard_key_, _OBJ_END_)
			if(shard_key == _OBJ_END_):
				# querying on secondary shards as main shard key doesn't exist
				# so try each secondary shard and test if we can query on it
				# otherwise we go with primary shard
				for secondary_shard_key_name, _shard in cls._secondary_shards_.items():
					secondary_shard_key = _query.get(secondary_shard_key_name, _OBJ_END_)
					if(secondary_shard_key == _OBJ_END_):
						continue
					#check if we can construct a secondary shard query
					#extract keys from the query to create new secondary shard query
					not_possible_in_the_secondary_shard = False
					if(sort):
						#check if we can perform sort on the secondary index
						for sort_key, sort_direction in sort:
							if(sort_key not in _shard.attributes):
								#cannot perform sort
								not_possible_in_the_secondary_shard = True

					secondary_shard_query = {}
					for query_attr_name, query_attr_val in _query.items():
						attr_exists_in_shard = _shard.attributes.get(query_attr_name, _OBJ_END_)
						if(attr_exists_in_shard == _OBJ_END_):
							not_possible_in_the_secondary_shard = True
							break
						secondary_shard_query[query_attr_name] = query_attr_val

					if(not_possible_in_the_secondary_shard):
						#try next secondary shard
						continue

					secondary_collection_shards = _shard._Model_.get_collections_from_shard_key(secondary_shard_key)
					for collection_shard in secondary_collection_shards:
						_key = id(collection_shard)
						shard_and_queries = collections_to_query.get(_key)
						if(shard_and_queries == None):
							collections_to_query[_key] \
								= shard_and_queries \
								= (collection_shard, [], _shard._Model_)
						shard_and_queries[1].append(secondary_shard_query)
					break

			if(shard_key != _OBJ_END_ and not_possible_in_the_secondary_shard):
				#query on the primary shard
				collection_shards = cls.get_collections_from_shard_key(shard_key)
				for collection_shard in collection_shards:
					_key = id(collection_shard)
					shard_and_queries = collections_to_query.get(_key)
					if(shard_and_queries == None):
						collections_to_query[_key]\
							= shard_and_queries \
							= (collection_shard, [], cls)
					shard_and_queries[1].append(_query)

		#if we did not find any possible shard to query
		# we query all primary shards and assemble queries using chain
		# which will be dead slow and possibly fucked up!
		if(not collections_to_query):
			collections_to_query = {id(x): (x, queries, cls) for x in cls.get_collection_on_all_db_nodes()}

		def count_documents(_collection, _query, offset, limit):
			kwargs = {}
			if(offset):
				kwargs["skip"] = offset
			if(limit):
				kwargs["limit"] = limit
			return _collection.count_documents(_query, **kwargs)

		def query_collection(_Model, _collection, _query, offset=None):

			IS_DEV and MONGO_DEBUG_LEVEL > 1 and print(
				"#MONGO: querying", _Model, _collection, _query, sort, offset, limit
			)

			ret = _collection.find(_query, **kwargs)
			if(sort): # from global arg
				ret = ret.sort(sort)
			if(offset): # from global arg
				ret = ret.skip(offset)
			if(limit):
				ret = ret.limit(limit)


			# we queried from the secondary shard, will not have all fields
			if(_Model._is_secondary_shard and not _no_requery):
				# do a requery to fetch full document
				#TODO: make it batch wise fetch
				def batched_requery_iter(ret, n=200):
					for docs in batched_iter(ret, n):
						_query = {"$or": [dict(cls.pk_from_doc(_doc)) for _doc in docs]}
						IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: requerying", cls, _query)
						for item in cls.query(_query):
							yield item
				ret = batched_requery_iter(ret)
			else:
				ret = map(cls.get_instance_from_document, ret)

			return ret, lambda: count_documents(_collection, _query, offset, limit)

		threads = []
		for _collection_shard_id, shard_and_queries in collections_to_query.items():
			_collection_shard, _queries, _Model = shard_and_queries
			new_query = None
			if(len(_queries) == 1):
				new_query = _queries[0]
			else:
				new_query = {"$or": _queries}

			if(offset and len(collections_to_query) > 1):
				offset = None
				print("Warning: Cannot use offset when query spans multiple collections shards, improve your query")

			thread = gevent.spawn(
				query_collection,
				_Model,
				_collection_shard,
				new_query,
				offset=offset
			)
			threads.append(thread)
							
		class SortKey(object):
			def __init__(self, obj):
				self.obj = obj

			def __lt__(self, other_obj):
				if(not sort):
					return True
				for sort_key, sort_direction in sort:
					value1 = getattr(self.obj, sort_key, None)
					value2 = getattr(other_obj, sort_key, None)
					if(sort_direction > 0):
						if(value1 == value2):
							continue
						if(value1 == None or value2 == None):
							return True if not value1 else False
						return value1 < value2
					else:
						if(value1 == value2):
							continue
						if(value1 == None or value2 == None):
							return False if not value1 else True
						return value1 > value2
				return True

		#does local sorting on results from multiple query iterators using a heap
		class MultiCollectionQueryResult:
			query_result_iters = None
			query_count_iters = None
			buffer = None

			def __init__(self):
				self.query_result_iters = []
				self.query_count_iters = []

			def add(self, query_result_iter, query_count_iter=None):
				self.query_result_iters.append(query_result_iter)
				if(query_count_iter):
					self.query_count_iters.append(query_count_iter)

			def count(self):
				ret = 0
				for count_func in self.query_count_iters:
					ret += count_func()
				return ret

			def push_query_iter_into_heap(self, _query_result_iter):
				try:
					next_value = next(_query_result_iter)
					heapq.heappush(self.buffer, (SortKey(next_value), _query_result_iter, next_value))
				except StopIteration:
					pass

			def __iter__(self):
				if(len(self.query_result_iters) == 1):
					return self.query_result_iters[0]
				#else we sort results for each if there is a sorting key given
				self.buffer = []
				heapq.heapify(self.buffer)
				for _query_result_iter in self.query_result_iters:
					self.push_query_iter_into_heap(_query_result_iter)
				return self

			def __next__(self):
				_ret = None
				try:
					sort_value, _query_result_iter, _ret = heapq.heappop(self.buffer)
					self.push_query_iter_into_heap(_query_result_iter)
				except IndexError:
					raise StopIteration

				return _ret


		#waint on threads and return cursors
		multi_collection_query_result = MultiCollectionQueryResult()

		for thread in gevent.joinall(threads):
			multi_collection_query_result.add(thread.value[0], thread.value[1])

		return multi_collection_query_result

	'''
		usually we call this while compile time,
		perfomance wise is not an issue
	'''
	@classmethod
	def on(cls, events, func):
		if(not isinstance(events, list)):
			events = [events]
		for event in events:
			event_listeners = getattr(cls, "_event_listeners_", _OBJ_END_)
			if(event_listeners == _OBJ_END_):
				event_listeners = {}
				setattr(cls, "_event_listeners_", event_listeners)
			handlers = event_listeners.get(event)
			if(handlers == None):
				event_listeners[event] = handlers = []
			handlers.append(func)

	@classmethod
	def _trigger_event(cls, event, obj):
		handlers = cls._event_listeners_.get(event)
		if(not handlers):
			return
		for handler in handlers:
			handler(obj)

	def before_update(self):
		pass

	def commit(self, force=False, ignore_exceptions=None):
		cls = self.__class__
		committed = False

		if(self.__is_new): # try inserting
			cls._trigger_event(EVENT_BEFORE_CREATE, self)
			if(not self._set_query_updates):
				return self # nothing to update
			shard_key_name = cls._shard_key_
			if(shard_key_name == "_id" and self._id == None):
				raise Exception("Need to sepcify _id when sharded by _id")

			IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: new object values", cls, self._set_query_updates)
			self.__is_new = False
			_collection_shard = cls.get_collection(
				str(getattr(self, shard_key_name))
			)
			try:
				#find which shard we should insert to and insert into that
				self._insert_result = _collection_shard.insert_one(
					self._set_query_updates
				)

				_id = self._insert_result.inserted_id
				# try inserting into secondary shards now
				for shard_key, _shard in cls._secondary_shards_.items():
					shard_key_value = self._set_query_updates.get(shard_key)
					if(shard_key_value == None):
						#shard key doesn't exist
						continue

					_secondary_insert_values = {}
					for attr in _shard.attributes:
						_attr_val = self._set_query_updates.get(attr, _OBJ_END_)
						if(_attr_val != _OBJ_END_):
							_secondary_insert_values[attr] = _attr_val
					#use the same id of the document
					_secondary_insert_values["_id"] = _id
					#insert into other shards
					try:
						_shard._Model_.get_collection(
							str(getattr(self, shard_key))
						).insert_one(
							_secondary_insert_values
						)
					except Exception as ex:
						print("#MONGO: encountered an error while updating secondary shard:", shard_key, str(ex))
						#delete the original insert
						_collection_shard.delete_one({"_id": _id})
						raise ex


				#set _id updates
				self._id = self._set_query_updates["_id"] = _id
				committed = True
				# set original doc and custom dict and set fields
				#copy the dict to another
				self.reinitialize_from_doc(dict(self._set_query_updates))
				#hook to do something for newly created db entries
				cls._trigger_event(EVENT_AFTER_CREATE, self)

			except DuplicateKeyError as ex:

				if(ignore_exceptions and DuplicateKeyError in ignore_exceptions):
					return None

				if(not force):
					raise(ex) # re reaise

				#get original doc from mongo shard
				# and update any other fields
				self.reinitialize_from_doc(
					_collection_shard.find_one(self.pk())
				)
				IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: created a duplicate, refetching and updating",
					self.pk()
				)

				# try removing all primary keys
				# although this unnecessary it's good
				# to mongo ?
				if("_id" in self._set_query_updates):
					del self._set_query_updates["_id"]

				for k in list(self._set_query_updates.keys()):
					if(k in cls._pk_attrs):
						del self._set_query_updates[k]



		if(not self.__is_new and not committed):  # try updating
			_update_query = {}
			if(self._set_query_updates):
				_update_query["$set"] = self._set_query_updates
			if(self._other_query_updates):
				_update_query.update(self._other_query_updates)
			if(not _update_query and not force):
				return self # nothing to update

			self.update(_update_query) # , hint=self.__class__._pk_attrs)


		#clear and reset pk to new
		self.pk(renew=True)
		#clear
		self._set_query_updates.clear()
		self._other_query_updates.clear()

		return self

	def delete(self):
		_Model = self.__class__
		if(_Model._is_secondary_shard):
			raise Exception("Cannot delete secondary shard item")
		#Note: when we know the _id and the shard we basically delete them by _id

		_Model._trigger_event(EVENT_BEFORE_DELETE, self)

		#delete it from secondary shards first
		for _shard_key, _seconday_shard in _Model._secondary_shards_.items():
			_seconday_Model = _seconday_shard._Model_
			collection_shard = _seconday_Model.get_collection(
				getattr(self, _shard_key)
			)
			_delete_query = {"_id": self._id}
			collection_shard.delete_one(_delete_query)
			IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: deleting from secondary",
				_seconday_Model, _delete_query
			)

		#find which pimary shard it belongs to and delete it there
		collection_shard = _Model.get_collection(
			str(getattr(self, _Model._shard_key_))
		)
		_delete_query = {"_id": self._id}
		collection_shard.delete_one(_delete_query)
		IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: deleting from primary",
				_Model, _delete_query
			)
		if(_Model.__cache__):
			_Model.__cache__.delete(self.pk_tuple())

		_Model._trigger_event(EVENT_AFTER_DELETE, self)


	@classmethod
	def remove_from_cache(cls, *_objs):
		if(not cls.__cache__):
			return
		for _obj in _objs:
			cls.__cache__.delete(_obj.pk_tuple())


class SecondaryShard:
	attributes = None
	indexes = None
	collection_name = None
	_Model_ = None

	def __init__(self):
		self.attributes = {}
		self.indexes = []


#control
class CollectionTracker(Model):
	_db_name_ = "control"
	_collection_name_ = "collection_tracker"

	_id = Attribute(str) # db_name__collection_name
	db_nodes = Attribute(list)
	primary_shard_key = Attribute(str)
	secondary_shard_keys = Attribute(list)


#tags shard keys to the attributes and use it when intializing the model
def SHARD_BY(primary=None, secondary=None, Model=None):
	if(primary and isinstance(primary, Attribute)):
		primary.is_primary_shard_key = True
	if(secondary):
		for secondary_shard_attr in secondary:
			if(secondary_shard_attr and isinstance(secondary_shard_attr, Attribute)):
				secondary_shard_attr.is_secondary_shard_key = True

#tags indexes_to_create to attributes and retrieve them when initializing the model
def INDEX(*indexes):
	for index_key_set in indexes:
		first_key_of_index_key_set = index_key_set[0]
		if(isinstance(first_key_of_index_key_set, tuple)): # when it has a sorting order as second key
			first_key_of_index_key_set = first_key_of_index_key_set[0]

		first_key_of_index_key_set._indexes_to_create = getattr(first_key_of_index_key_set, "_indexes_to_create", [])
		first_key_of_index_key_set._indexes_to_create.append(index_key_set)


'''
Control Jobs
- 	every time a job is completed, check if it has parent,
	goto parent and check if all child jobs are finished and mark it complete
'''
class ControlJobs(Model):
	_db_name_ = "control"
	_collection_name_ = "jobs"

	parent__id = Attribute(str)
	num_child_jobs = Attribute(int, default=0) # just for reconcillation
	_type = Attribute(int) # reshard=>0, create_secondary_index=>1, create_primary_index=>2
	db = Attribute(str)
	collection = Attribute(str)
	status = Attribute(int, default=0) # 0=>not started, 1=>progress, 2=>completed
	uid = Attribute(str) # unique identifier not to duplicate jobs

	worker_id = Attribute(str) # worker id
	worker_should_update_within_ms = Attribute(int, default=60000) # contract that worker should update every few millis
	#work data
	data = Attribute(dict)
	data1 = Attribute(dict)
	created_at = Attribute(int, default=cur_ms)
	updated_at = Attribute(int, default=cur_ms)

	INDEX(
		(db, collection, _type, uid),
		(db, collection, _type, status, {"unique": False}),
		(parent__id, {"unique": False})
	)

	SHARD_BY(primary=db)



	##Constants
	CREATE_SECONDARY_SHARD = 1

	def before_update(self):
		self.updated_at = cur_ms()


_cached_mongo_clients = {}
#DatabaseNode is basically a server
class DatabaseNode:

	_cached_pymongo_collections = {}
	#mongo connection
	mongo_connection = None
	hosts = None
	replica_set = None
	username = None
	password = None
	#default db name
	db_name = None

	#this initializes the tables in all nodes
	def __init__(self, hosts=None, replica_set=None, username=None, password=None, db_name=None):
		if(isinstance(hosts, str)):
			hosts = [hosts]
		hosts.sort()
		self.hosts = hosts
		self.replica_set = replica_set
		self.username = username
		self.db_name = db_name
		# use list of hosts, replicaset and username as cache key
		# so as not to create multiple clients
		_mongo_clients_cache_key = (",".join(hosts), replica_set or "" , username or "")
		self.mongo_connection = _cached_mongo_clients.get(_mongo_clients_cache_key)
		if(not self.mongo_connection):
			self.mongo_connection = _cached_mongo_clients[_mongo_clients_cache_key] = MongoClient(host=hosts, replicaSet=replica_set, username=username, password=password)


	def to_dict(self):
		return {
			"hosts": self.hosts,
			"replica_set": self.replica_set,
			"username": self.username,
			"db_name": self.db_name
		}

	#returns collection on the given DatabaseNode
	def get_collection(self, Model):
		ret = self._cached_pymongo_collections.get(Model)
		if(not ret):
			self._cached_pymongo_collections[Model] \
				= ret \
				= self.mongo_connection[Model._db_name_][Model._collection_name_with_shard_]
		return ret


def initialize_model(Model):

	if(isinstance(Model, list)):
		for _m in Model:
			initialize_model(_m)
		return

	#defaults
	Model._shard_key_ = getattr(Model, "_shard_key_", "_id")
	Model._secondary_shards_ = {}
	Model._indexes_ = getattr(Model, "_indexes_", [("_id",)])

	# temp usage _id_attr
	_id_attr = Attribute(str) # default is of type objectId
	Model._attrs = _model_attrs = {"_id": _id_attr}
	Model._pk_attrs = None
	'''it's used for translating attr objects/name to string names'''
	Model._attrs_to_name = attrs_to_name = {_id_attr: '_id', '_id': '_id'}
	for k, v in Model.__dict__.items():
		if(isinstance(v, Attribute)):
			_model_attrs[k] = v
			attrs_to_name[v] = k
			# dumb , but it's one time thing
			# and also helps converting if any given attributes as strings
			attrs_to_name[k] = k

			#check if it has any shard, then extract indexes, shard key tagged to attributes
			if(not Model._is_secondary_shard): # very importat check
				#check indexes_to_create
				_indexes_to_create = getattr(v, "_indexes_to_create", None)
				if(_indexes_to_create):
					delattr(v, "_indexes_to_create") # because we don't need it again after first time
					Model._indexes_.extend(_indexes_to_create)

				is_primary_shard_key = getattr(v, "is_primary_shard_key", False)
				is_secondary_shard_key = getattr(v, "is_secondary_shard_key", False)
				if(is_primary_shard_key is True):
					Model._shard_key_ = k
					delattr(v, "is_primary_shard_key") # because we don't need it again after first time
				elif(is_secondary_shard_key is True):
					Model._secondary_shards_[k] = SecondaryShard()
					delattr(v, "is_secondary_shard_key") # because we don't need it again after first time

	IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: Initializing Model", Model,
		Model._indexes_, Model._shard_key_, Model._secondary_shards_
	)

	Model._event_listeners_ = getattr(Model, "_event_listeners_", {})
	#some default event
	Model.on(EVENT_BEFORE_UPDATE, lambda obj: obj.before_update())

	_pymongo_indexes_to_create = []
	Model._pk_is_unique_index = False
	for _index in Model._indexes_:
		mongo_index_args = {}
		pymongo_index = []
		if(not isinstance(_index, tuple)):
			_index = (_index,)
		for _a in _index:
			_attr_name = _a
			_ordering = pymongo.ASCENDING
			if(isinstance(_a, tuple)):
				_a, _ordering = _a
			if(isinstance(_a, Attribute)):
				_attr_name = attrs_to_name[_a]
			if(isinstance(_a, str)):
				_attr_name = _a
			if(isinstance(_a, dict)):
				mongo_index_args = _a
				continue

			pymongo_index.append((_attr_name, _ordering))

		#convert to tuple again
		pymongo_index = tuple(pymongo_index)
		is_unique_index = mongo_index_args.get("unique", True) is not False

		_index_shard_key = pymongo_index[0][0] # shard key is the first mentioned key in index

		if(Model._shard_key_ != _index_shard_key):
			#secondary shard tables
			_seconday_shard = Model._secondary_shards_.get(_index_shard_key)

			if(_seconday_shard):
				for _attr_name, _ordering in pymongo_index:
					_seconday_shard.attributes[_attr_name] = getattr(
						Model,
						_attr_name
					)
				#create _index_ for seconday shards
				_secondary_shard_index = list(pymongo_index)
				_secondary_shard_index.append(mongo_index_args)
				_seconday_shard.indexes.append(
					tuple(_secondary_shard_index)
				)

		ignore_index_creation = False
		#check if index it belongs to primary shard
		if(	Model._shard_key_ == _index_shard_key):
			#index belong to primary shard
			if(not Model._pk_attrs
				or (is_unique_index and not Model._pk_is_unique_index)
			):
				Model._pk_is_unique_index = is_unique_index
				Model._pk_attrs = _pk_attrs = OrderedDict()
				for i in pymongo_index: # first unique index
					_pk_attrs[i[0]] = 1
		elif(_index_shard_key in Model._secondary_shards_):
			#this index goes into the secondary table`
			#Ex: Table A is sharded by x then we created indexes that has x on A_shard_x table only
			ignore_index_creation = True

		#check for indexes to ignore
		if(pymongo_index[0][0] == "_id"):
			ignore_index_creation = True

		#create the actual index
		if(not ignore_index_creation):
			_pymongo_indexes_to_create.append((pymongo_index, mongo_index_args))


	if(not Model._pk_attrs): # create default _pk_attrs
		Model._pk_attrs = OrderedDict(_id=True)

	# set collection name to include shard_keys
	Model._collection_name_with_shard_ = Model._collection_name_ + "_shard_" + Model._shard_key_

	##find tracking nodes
	Model._collection_tracker_key_ = "%s__%s"%(Model._db_name_, Model._collection_name_with_shard_)

	IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO collection tracker key", Model._collection_tracker_key_)

	if(Model not in [CollectionTracker, ControlJobs]):
		collection_tracker = CollectionTracker.get(Model._collection_tracker_key_)
		if(not collection_tracker
			or not collection_tracker.db_nodes
			or not collection_tracker.primary_shard_key
		):
			print(
				"#MONGOORM_IMPORTANT_INFO : "
				"Collection tracker entry not present for '%s'.."
				"creating table in control node for this time in database '%s'. "
				"You may want to talk to dba to move to a proper node"%(Model.__name__, Model._db_name_)
			)

			#check if the Model already has _db_nodes_
			db_nodes = getattr(Model, "_db_nodes_", None) or [
						{
							"hosts": init_db_nodes.hosts,
							"replica_set": init_db_nodes.replica_set,
							"username": init_db_nodes.username,
							"db_name": Model._db_name_
						} for init_db_nodes in CollectionTracker._db_nodes_
					]

			collection_tracker = CollectionTracker(
				_id=Model._collection_tracker_key_,
				db_nodes=db_nodes,
				primary_shard_key=Model._shard_key_,
				secondary_shard_keys=list(Model._secondary_shards_.keys())
			).commit(force=True)

		Model._db_nodes_ = [DatabaseNode(**_db_node) for _db_node in collection_tracker.db_nodes]
		#TODO: find new secondary shards by comparing collection_tracker.secondary_shard_keys, Model._secondary_shards_.keys()
		#and create a job to create and reindex all data to secondary index
		if(collection_tracker.primary_shard_key != Model._shard_key_):
			raise Exception("#MONGO_EXCEPTION: Primary shard key changed for ", Model)

		#create diff jobs
		to_create_secondary_shard_key, to_delete_secondary_shard_key = list_diff2(
			Model._secondary_shards_.keys(),
			collection_tracker.secondary_shard_keys
		)
		for shard_key in to_create_secondary_shard_key:
			try:
				ControlJobs(
					db=Model._db_name_,
					collection=Model._collection_name_with_shard_,
					_type=ControlJobs.CREATE_SECONDARY_SHARD,
					uid=shard_key
				).commit()
				IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: create a control job to create secondary sharding", Model, shard_key)
			except DuplicateKeyError as ex:
				pass

	#TODO: create or delete index using control jobs
	for pymongo_index, additional_mongo_index_args in _pymongo_indexes_to_create:
		mongo_index_args = {"unique": True}
		mongo_index_args.update(**additional_mongo_index_args)

		IS_DEV and MONGO_DEBUG_LEVEL > 1 and print("#MONGO: creating_indexes", Model, pymongo_index, mongo_index_args)
		#in each node create indexes
		for db_node in Model._db_nodes_:
			db_node.get_collection(Model).create_index(pymongo_index, **mongo_index_args)
	
	#create secondary shards
	for _seconday_shard_key, _seconday_shard in Model._secondary_shards_.items():
		if(not Model._pk_is_unique_index):
			raise Exception("Cannot have secondary shard keys for non unique indexes! %s"%(Model,))

		class_attrs = {
			"_indexes_": _seconday_shard.indexes,
			"_shard_key_": _seconday_shard_key,
			"_collection_name_": Model._collection_name_,
			"_is_secondary_shard": True,
			"_db_nodes_": None
		}
		# add primary key attributes of the main class to
		# secondary shards too

		for attr_name in Model._pk_attrs:
			_seconday_shard.attributes[attr_name] = getattr(Model, attr_name)

		secondary_id_attr = getattr(Model, '_id', None)
		if(secondary_id_attr):
			_seconday_shard.attributes['_id'] = secondary_id_attr

		class_attrs.update(_seconday_shard.attributes)
		#also include the primary key of the primary shard
		#into secondary shards

		_seconday_shard._Model_ = type(
			"%s_%s"%(Model.__name__, _seconday_shard_key.upper()),
			(Model,),
			class_attrs
		)
		#initialize this new model
		initialize_model(_seconday_shard._Model_)


#initialize control Tr
#initialize all other nodes
def initialize_mongo(init_db_nodes, default_db_name=None):

	default_db_name = default_db_name or "temp_db"
	#initialize control db
	if(isinstance(init_db_nodes, dict)):
		init_db_nodes = DatabaseNode(**init_db_nodes)
	#check connection to mongodb
	init_db_nodes.mongo_connection.server_info()

	#initialize control db
	CollectionTracker._db_nodes_ = [init_db_nodes]
	ControlJobs._db_nodes_ = [init_db_nodes]
	initialize_model([CollectionTracker, ControlJobs])

	#set default db name for each class
	for cls in Model.__subclasses__():
		if(not getattr(cls, "_db_name_", None)):
			cls._db_name_ = default_db_name
		initialize_model(cls)



# Initially intended to be < 500 line and crazy scalable like
# add more nodes and it automatically shards
# - Complete client side sharding, control database,
# - You can Configure mongo as a sharded cluster.Just pass the configuration server to
# - init_mongo_cluster({"uri": "mongodb://localhost:27017", "db_name": "yourdb"})

#reshard-restart-reshard
